""" Module containing the Data class"""
# -*- coding: utf-8 -*-

from .base_file import GENEfile
import numpy as np

class Data:
    """ Class to provide the data containers of a GENE parameters/simulation"""
    def __init__(self, in_folder, extensions, runs=None):  
        self.av_vars = {'field': False,
                        'mom': False}
    
        self.av_times={'field': None,
                        'mom': None}
        self.field=None
        self.mom=[]
        
        if runs:
            if not isinstance(runs, list):
                runs = [runs]
            if not isinstance(extensions, list):
                extensions = [extensions]
            for run,ext in zip(runs,extensions):
                self.__map_a_run(in_folder, ext, run.parameters)
                
                    
    def __map_a_run(self, folder, extension, parameters):
        """
        for each field that has been saved in this restart we
        1. create a file object if needed,
        2. update the available variables
        3. get the times
        """
        # field file
        if parameters.pnt.istep_field > 0:
            if not self.av_vars['field']:
                self.field=GENEfile(folder, extension, 'field', parameters)
                self.av_vars['field']=True
                t,s,f=self.field.get_times_and_inds()
                self.av_times['field']=TimeStep(t,s,f)
            else:    
                self.field.redirect(folder,extension, parameters)
                t,s,f=self.field.get_times_and_inds()
                self.av_times['field'].join_continuation(t,s,f)
                
        # mom files
        if parameters.pnt.istep_mom > 0:
            if not self.av_vars['mom']:
                self.av_vars['mom']=True
                for i_spec in np.arange(parameters.pnt.n_spec):
                    self.mom.append(GENEfile(folder, extension, 'mom', parameters,parameters.species[i_spec]['name']))
                
                t,s,f=self.mom[0].get_times_and_inds()
                self.av_times['mom']=TimeStep(t,s,f)
            else:    
                self.field.redirect(folder,extension,parameters)
                t,s,f=self.field.get_times_and_inds()
                self.av_times['mom'].join_continuation(t,s,f)
   


class AvailableTimes:
    pass


class TimeStep:
    def __init__(self, times, steps, files):
        self.times = times  # times in GENE units
        self.steps = steps  # step in the file 
        self.files = files  # file extension      
        
    def join_continuation(self, times, steps, files):
        self.steps = np.append(self.steps[self.times<times[0]], steps)
        for i, e in reversed(list(enumerate(self.files))):
            if not self.times[i] < times[0]:
                self.files.remove(e)
            else:
                break
        self.files.extend(files)
        self.times = np.append(self.times[self.times<times[0]], times)
        



# -*- coding: utf-8 -*-

from bisect import bisect_left, bisect_right
from copy import deepcopy
import numpy as np


class Loader:
    """ This does the nasty job. We need for a given timestep to be able to load
        any kind of data. The idea is the following:
            1- the user sets folder and runs
            2- GUI determines which files and variables are available
            3- GUI fetches all times from all the files
            4- user selects a time window and a series of diagnostics
            5- GUI takes the subset of files that are needed
            5- GUI takes the requested subinterval in time
            6- loop over all times

            -- I win.
            """

    def __init__(self):
        pass
    
    def set_interval(self, diagnostics, data, run, t_start, t_end, step):    
        self.need_file={'field': False,
                        'mom': False}

        #print("in loader data.av_times['field'].times",data.av_times['field'].times)
        #dummy = input('press key')
        #trigger the file needed from all diagnostic        
        for diag in diagnostics:
            my_files=diag.get_info()
            for k in self.need_file.keys():
                self.need_file[k]=self.need_file[k] or my_files[k]
            diag.setup_options(run)
        
        #append all times together
        #TODO this is bad. shoould rather use istep here to simplify
        times=np.array([])
        for k in self.need_file.keys():
            if self.need_file[k]:
                if times.size==0:
                    times=data.av_times[k].times
                else:
                    times=np.intersect1d(times, data.av_times[k].times)
                
        times=np.unique(times)
        
        i_s = bisect_left(times, t_start)
        i_e = bisect_right(times, t_end)
        if i_s == len(times):
            i_s -= 1
        self.times = times[i_s:i_e:step]
        self.t_start = times[0]
        self.t_end = times[-1]
        
        self.steps=[{}]*self.times.size
        self.files=[{}]*self.times.size

        #TODO extension is the same, no need of dictionary
        for k in self.need_file.keys():
            if self.need_file[k]:
                dum, dum2, inds=np.intersect1d(self.times,data.av_times[k].times, return_indices=True)
                for i_st,st in enumerate(inds):
                
                    if not self.steps[i_st]:
                        self.steps[i_st]={k: data.av_times[k].steps[st]} 
                        self.files[i_st]={k: data.av_times[k].files[st]} 
                    else:
                        self.steps[i_st].update({k: data.av_times[k].steps[st]})
                        self.files[i_st].update({k: data.av_times[k].files[st]}) 
                



"""Module containing the treatment of GENE's geometry output"""
import numpy as np
import h5py
from .par_io import Parameters
import os
import matplotlib.pyplot as plt


class Geometry:
    """ Class to handle geometry input from GENE runs
        parameters: either Parameters object or parameters as a namedtuple
    """
   
    def __init__(self,folder,extension,parameters):
        ##self.pnt is the parameter dictionary as a namedtuple
        if isinstance(parameters, Parameters):
            self.pnt = parameters.asnamedtuple()
        elif isinstance(parameters, tuple):
            self.pnt = parameters

        self.geomtype = self.pnt.magn_geometry

        ##get geometry from file
        if self.pnt.write_h5:
            self.getgeom_h5(folder,extension)
        elif self.pnt.write_adios:
            raise NotImplementedError("ADIOS not supported")
        else:
            self._getgeom_std(folder,extension)
            
        self.t=1
        ##set plot attribute 
        if self.pnt.x_local:
            setattr(self, 'plot', self.__plot_geom_local)
        if self.pnt.x_global:
            setattr(self, 'plot', self.__plot_geom_xglobal)
        if self.pnt.is3d:
            setattr(self, 'plot', self.__plot_geom_xyglobal)
        
    def _getgeom_std(self,folder,extension):
        geom = self.getgeom(folder,extension)
        if not self.pnt.x_local:
            self.jaco3d = self.jacobian[:,np.newaxis,:]
        elif not self.pnt.y_local:
            raise NotImplementedError("y_local not supported")
        return geom

    @staticmethod
    def untangle_1d(arrlist, start, nxorny):
        """ Process a geometry file section that contains a 1d (x or y) field

        :param arrlist: The entire file as a list of lines
        :param start: The starting index of the quantity in arrlist
        :param nxorny: nx0 or nky0 depending on the type of simulation
        :returns: A 1d numpy array
        """
        arr_1d = np.zeros(nxorny)
        i = 0
        for line in arrlist[start:-1]:
            for j in range(len(line.split())):
                arr_1d[i + j] = line.split()[j]
            i += len(line.split())
            if i >= nxorny:
                break
        return arr_1d

    @staticmethod
    def untangle_2d(arrlist, start, nxorny, nz):
        """ Process a geometry file section that contains a 2d (x or y, z) field

        :param arrlist: The entire file as a list of lines
        :param start: The starting index of the quantity in arrlist
        :param nxorny: nx0 or nky0 depending on the type of simulation
        :param nz: nz0, number of z grid points
        :returns: A 2d numpy array
        """
        arr_2d = np.zeros((nz, nxorny))
        ik = 0
        for line in arrlist[start:-1]:
            arr = line.split()[:]
            lb = ik % nxorny
            ub = min(lb + 16, nxorny)
            k = int(ik/nxorny)
            dim = ub - lb
            arr_2d[k, lb:ub] = arr[0:dim]
            if dim < 16:
                arr_2d[k + 1, 0:16 - dim] = arr[dim:]
            ik += 16
            if ik >= nz*nxorny:
                break
        return arr_2d.T         # want (nx,nz) order

    def getgeom(self,folder,extension,):
        """ Returns the geometry from a non-hdf5 file """
        local = self.pnt.x_local

        with open(os.path.join(folder, "{}{}".format(self.geomtype.strip("'"), \
                                                extension)), "r") as geomfile:
            if local:
                geom = self.getgeom_loc(geomfile, self.pnt.nz0)
            elif self.pnt.y_local:  # x-global
                geom = self.getgeom_glob(geomfile, self.pnt.nz0, self.pnt.nx0)
            elif self.pnt.x_local:  # y-global
                raise NotImplementedError("x_local (y global) not supported")
            else:
                raise NotImplementedError("fully global not supported")
        return geom

    def getgeom_loc(self, geomfile, nz):
        geom = np.empty((16, nz), dtype=np.float64)
        k = 0
        for line in geomfile:
            if len(line.split()) == 16:
                geom[:, k] = line.split()[:]
                k += 1
            elif line.startswith('Cy'):
                self.Cy = float(line.split()[-1])
            elif line.startswith('Cxy'):
                self.Cxy = float(line.split()[-1])
            elif line.startswith('q0'):
                self.q = float(line.split()[-1])
        self.gxx = geom[0]
        self.gxy = geom[1]
        self.gxz = geom[2]
        self.gyy = geom[3]
        self.gyz = geom[4]
        self.gzz = geom[5]
        self.Bfield = geom[6]
        self.dBdx = geom[7]
        self.dBdy = geom[8]
        self.dBdz = geom[9]
        self.jacobian = geom[10]
        self.R = geom[11]
        self.Z = geom[13]
        self.dxdR = geom[14]
        self.dxdZ = geom[15]
        gamma1 = self.gxx*self.gyy - self.gxy**2
        gamma2 = self.gxx*self.gyz - self.gxy*self.gxz
        gamma3 = self.gxy*self.gyz - self.gyy*self.gxz
        self.Kx = -self.dBdy - gamma2/gamma1*self.dBdz
        self.Ky = self.dBdx - gamma3/gamma1*self.dBdz
        if not hasattr(self, 'Cxy'):
            self.Cxy=1.0
        return geom

    def getgeom_glob(self, geomfile, nz0, nxorny):
        """ Subroutine for geometry files from global runs """
        NUMFIELDS = 19
        geom = []

        geomdict = {'q': 0, 'gxx': 1, 'gxy': 2, 'gxz': 3, 'gyy': 4, 'gyz': 5,
                    'gzz': 6, 'Bfield': 7, 'dBdx': 8, 'dBdy': 9, 'dBdz': 10,
                    'jacobian': 11, 'C_y': 12, 'C_xy': 13, 'geo_R': 14,
                    'geo_Z': 15, 'geo_c1': 16, 'geo_c2': 17, 'dpdx_pm_arr': 18}
        pos_start = np.zeros(NUMFIELDS, dtype=int)  # where a field starts
        num = 0
        parheader = True
        geomlist = geomfile.readlines()
        for linnum, line in enumerate(geomlist):
            if parheader:  # ignore the top part of the file (copy of the geometry namelist)
                if line.startswith(r'/'):
                    parheader = False
                continue
            if len(line.split()) == 1:  # the variable names
                try:
                    num = geomdict[line.strip()]
                    pos_start[num] = linnum + 1
                except KeyError:
                    try:  # Test if it is a single number (can occur for 1d arrays)
                        float(line.strip())
                    except ValueError:
                        raise RuntimeError("Unknown entry name in geometry file")
        for num in range(NUMFIELDS):
            if pos_start[num] == 0:  # This should only occur if a field does not exist
                continue
            if num in [0, 12, 13, 18]:  # We have a 1d field following
                tmp = self.untangle_1d(geomlist, pos_start[num], nxorny)
            else:
                tmp = self.untangle_2d(geomlist, pos_start[num], nxorny, nz0)
            geom.append(tmp)
        for key in geomdict:
            setattr(self, key, geom[geomdict[key]])
        self.Cy = self.C_y
        self.Cxy = self.C_xy
        self.q = self.q
        self.R = self.geo_R
        self.Z = self.geo_Z

        self.dxdR = self.geo_c1
        self.dxdZ = self.geo_c2
        del self.C_y
        del self.C_xy
        del self.geo_R
        del self.geo_Z
        del self.geo_c1
        del self.geo_c2
        self.dpdx_pm_arr
        return geom

    def getgeom_h5(self,folder,extension):
        """ Returns the geometry from a hdf5 file """

        geomfile = folder / "{}{}".format(self.geomtype.strip("'"), extension)
        geom = h5py.File(geomfile, 'r')

        self.gxx = geom.get('/metric/g^xx').value
        self.gxy = geom.get('/metric/g^xy').value
        self.gxz = geom.get('/metric/g^xz').value
        self.gyy = geom.get('/metric/g^yy').value
        self.gyz = geom.get('/metric/g^yz').value
        self.gzz = geom.get('/metric/g^zz').value
        self.Cy = geom.get('/metric/C_y').value
        self.Cxy = geom.get('/metric/C_xy').value

        self.Bfield = geom.get('/Bfield_terms/Bfield').value
        self.dBdx = geom.get('/Bfield_terms/dBdx').value
        self.dBdy = geom.get('/Bfield_terms/dBdy').value
        self.dBdz = geom.get('/Bfield_terms/dBdz').value
        self.jacobian = geom.get('/Bfield_terms/Jacobian').value

        if self.pnt.x_local:
            self.R = geom.get('/shape/R').value
            self.Z = geom.get('/shape/Z').value
            self.dxdR = geom.get('/shape/dxdR').value
            self.dxdZ = geom.get('/shape/dxdZ').value

        try:
            self.x_cart = geom.get('/cartesian_cords/x_cart').value
            self.y_cart = geom.get('/cartesian_cords/y_cart').value
            self.z_cart = geom.get('/cartesian_cords/z_cart').value
        except AttributeError:
            pass

        try:
            self.q = geom.get('/profile/q_prof').value
        except AttributeError:
            self.q = self.pnt.q0

        try:
            self.dpdx_arr = geom.get('/profile/dpdx_pm_arr').value
        except AttributeError:
            self.dpdx_arr = self.pnt.dpdx_pm
        geom.close()
        

    def __plot_geom_local(self, spatial_grid):
        self.names = {'gxx': r"$g^{xx}$", 'gxy': r"$g^{xy}$", 'gxz': r"$g^{xz}$",
                      'gyy': r"$g^{yy}$", 'gyz': r"$g^{yz}$", 'gzz': r"$g^{zz}$",
                      'dBdx': r"$dB/dx$", 'dBdy': r"$dB/dy$", 'dBdz': r"$dB/dz$",
                      'jacobian': r"$Jacobian$",
                      'dxdR': r"$dx/dR$", 'dxdZ': r"$dx/dZ$", }
        n_rows = 2
        n_cols = 8

        figure = plt.figure(figsize=(5, 5), dpi=100)
        i_plt = 0

        for fld, lbl in self.names.items():
            i_plt += 1
            a = figure.add_subplot(n_rows, n_cols, i_plt)
            a.plot(spatial_grid.z/np.pi, getattr(self, fld))
            a.set_title(lbl)
            a.set_xlabel(r'$z/\pi$')
            a.tick_params(axis='both', which='major', labelsize=8)
            a.tick_params(axis='both', which='minor', labelsize=8)

        """surface"""
        a = figure.add_subplot(n_rows, n_cols, i_plt+1)
        a.plot(self.R, self.Z)
        a.set_ylabel(r'$Z$ [m]')
        a.set_xlabel(r'$R$ [m]')
        a.tick_params(axis='both', which='major', labelsize=8)
        a.tick_params(axis='both', which='minor', labelsize=8)
        a.set_aspect('equal', adjustable='box')

        """k theta"""
        g_yy = (self.gxx * self.gzz -
                self.gxz**2) * self.jacobian**2
        g_yz = ((self.gxy * self.gxz -
                self.gxx * self.gyz) *
                self.jacobian**2)
        g_zz = ((self.gxx * self.gyy -
                self.gxy**2) * self.jacobian**2)
        e_theta = np.sqrt(
                (self.q * self.Cy)**2 * g_yy +
                2 * self.q * self.Cy * g_yz +
                g_zz)
        k2 = np.sqrt((self.gxx * self.gyy - self.gxy**2)/self.gxx)

        a = figure.add_subplot(n_rows, n_cols, i_plt+2)
        a.plot(spatial_grid.z, self.q * self.Cy / e_theta)
        a.plot(spatial_grid.z, k2)
        a.set_xlabel(r'$z/\pi$', fontsize=8)
        a.legend([r'$k_{\theta}$', r'$k_{2}$'])
        a.tick_params(axis='both', which='major')
        a.tick_params(axis='both', which='minor')
        return [figure]

    def __plot_geom_xglobal(self, spatial_grid=None):
        if not spatial_grid:
            raise ("Missing grids for plotting")
            
        self.names = {'gxx': r"$g^{xx}$", 'gxy': r"$g^{xy}$", 'gxz': r"$g^{xz}$",
                  'gyy': r"$g^{yy}$", 'gyz': r"$g^{yz}$", 'gzz': r"$g^{zz}$",
                  'dBdx': r"$dB/dx$", 'dBdy': r"$dB/dy$", 'dBdz': r"$dB/dz$",
                  'jacobian': r"$Jacobian$",
                  'dxdR': r"$dx/dR$", 'dxdZ': r"$dx/dZ$", }    
            
        n_rows = 2
        n_cols = 6

        figure_1 = plt.figure(figsize=(15, 15), dpi=100)
        figure_2 = plt.figure(figsize=(15, 15), dpi=100)
        i_plt = 0

        x, z = np.meshgrid(spatial_grid.x_a, spatial_grid.z)
        for fld, lbl in self.names.items():
            i_plt += 1

            a = figure_1.add_subplot(n_rows, n_cols, i_plt)
            a.pcolormesh(z, x, getattr(self, fld)[:, :], 
                         cmap=plt.get_cmap('jet'))
            a.set_title(lbl,)
            a.set_xlabel(r'$z/\pi$')
            a.set_ylabel(r'$x/a$')

            b = figure_2.add_subplot(n_rows, n_cols, i_plt)
            b.pcolormesh(self.R, self.Z, getattr(self, fld)[:, :], 
                         cmap=plt.get_cmap('jet'))
            b.set_aspect('equal', 'box')
            b.set_title(lbl)
            b.set_xlabel(r'$R$')
            b.set_xlabel(r'$Z$')

        """k theta"""
        g_yy = (self.gxx * self.gzz -
                self.gxz**2) * self.jacobian**2
        g_yz = ((self.gxy * self.gxz -
                self.gxx * self.gyz) *
                self.jacobian**2)
        g_zz = ((self.gxx * self.gyy -
                self.gxy**2) * self.jacobian**2)

        self.k_theta = np.zeros(self.gxx.shape, dtype=np.float64)
        for i_z in range(self.gxx.shape[0]):
            self.k_theta[i_z, :] =  self.q * self.Cy / np.sqrt(np.abs(     
                (self.q * self.Cy)**2 * g_yy[i_z, :] +
                2 * self.q * self.Cy * g_yz[i_z, :] +
                g_zz[i_z, :]))

        figure_3 = plt.figure(figsize=(5, 5), dpi=100)
        ax = figure_3.add_subplot(1, 1, 1)
        im = ax.pcolormesh(self.R, self.Z, self.k_theta, 
                      cmap=plt.get_cmap('jet'))
        figure_3.colorbar(im, ax=ax)
        ax.set_aspect('equal', 'box')
        ax.set_title(r'$k_{\theta}$')
        ax.set_xlabel(r'$R$')
        ax.set_xlabel(r'$Z$')
        
        figure_1.tight_layout()
        figure_2.tight_layout()
        figure_3.tight_layout()

        return [figure_1, figure_2, figure_3]

    def __plot_geom_xyglobal(self, sim):

        names = {'gxx': r"$g^{xx}$", 'gxy': r"$g^{xy}$", 'gxz': r"$g^{xz}$",
                 'gyy': r"$g^{yy}$", 'gyz': r"$g^{yz}$", 'gzz': r"$g^{zz}$",
                 'Bfield': "Bfield",
                 'dBdx': r"$dB/dx$", 'dBdy': r"$dB/dy$", 'dBdz': r"$dB/dz$",
                 'jacobian': r"$Jacobian$"}

        n_rows = 2
        n_cols = 6

        figure_1 = plt.figure(figsize=(5, 5), dpi=100)
        figure_2 = plt.figure(figsize=(5, 5), dpi=100)
        figure_3 = plt.figure(figsize=(5, 5), dpi=100)

        i_plt = 0

        for fld, lbl in names.items():

            x = sim.runs[0].spatialgrid.x_a
            y = sim.runs[0].spatialgrid.y
            z = sim.runs[0].spatialgrid.z/np.pi
            xind = sim.runs[0].spatialgrid.nx0
            yind = sim.runs[0].spatialgrid.ny0
            zind = sim.runs[0].spatialgrid.nz0

            i_plt += 1

            a = figure_1.add_subplot(n_rows, n_cols, i_plt)
            a.plot(z, getattr(self, fld)[:, int(yind/2), int(xind/2)])
            a.set_title(lbl, fontsize=8)
            a.set_xlabel(r'$z/\pi$', fontsize=8)

            a = figure_2.add_subplot(n_rows, n_cols, i_plt)
            a.plot(x, getattr(self, fld)[int(zind/2), int(yind/2), :])
            a.set_title(lbl, fontsize=8)
            a.set_xlabel(r'$x/a$', fontsize=8)

            a = figure_3.add_subplot(n_rows, n_cols, i_plt)
            a.plot(y, getattr(self, fld)[int(zind/2), :, int(xind/2)])
            a.set_title(lbl, fontsize=8)
            a.set_xlabel(r'$y/\rho_{ref}$', fontsize=8)

        figure_4 = plt.figure(figsize=(5, 5), dpi=100)

        a = figure_4.add_subplot(1, 2, 1)
        a.plot(sim.runs[0].spatialgrid.x_a, getattr(self, 'q'))
        a.set_title('q', fontsize=8)
        a.set_xlabel(r'$x/a$', fontsize=8)

        a = figure_4.add_subplot(1, 2, 2)
        a.plot(sim.runs[0].spatialgrid.x_a, getattr(self, 'dpdx_arr'))
        a.set_title('dpdx_arr', fontsize=8)
        a.set_xlabel(r'$x/a$', fontsize=8)

        return [figure_1, figure_2, figure_3, figure_4]



import numpy as np
import warnings
import scipy.interpolate as interp
import matplotlib.pyplot as plt
import matplotlib.tri as mtri
import matplotlib.animation as animation

import ..utils.comm
import ..utils.geom
import ..utils.averages
import ..data.slices
import .baseplot

import multiprocessing


class TorusContourCut(pydiag.data.slices.MomFieldSlice):
    """ Generate a phi (toroidal angle) cut of mom or field data and a triangulation """

    def __init__(self, common, species, rundatafiles, torangle=0, z_res=400, quantity="phi"):
        """ Constructor for torus diagnostic specific slice

        :param common: A CommonData object for the run to diagnose
        :param species: Name of the species to diagnose
        :param rundatafiles: RunDataFiles object, matching common
        :param torangle: The toroidal angle of the cut, phi=[0, 2*pi]
        :param z_res: The number of points that z should be interpolated to
        :param quantity: Name of the quantity to plot (from field or mom, refer to MomFieldSlice
        constructor
        """
        diagspace = pydiag.utils.comm.DiagSpace(common.spatialgrid, False, False, False,
                                                (None, None), (None, None), (None, None), False,
                                                False, False)
        super().__init__(common, quantity, species, diagspace, rundatafiles, lambda dum: dum)
        if common.pnt.magn_geometry == "miller":
            warnings.warn(
                    "TorusContourCut class lacks a correction factor for Miller geomtery")  # See
            #  Cyq0_r0 in IDL diag torus.pro

        self.torangle = torangle%(2*np.pi)
        self.species = species
        self.nz_extended = z_res
        self.z_full = np.append(self.cm.spatialgrid.z, np.pi)  # Extend the z grid to pi
        self.z_full_fine = np.linspace(self.z_full[0], self.z_full[-1], self.nz_extended)
        tortri = TorusCutTriangulation(common, z_res)
        self.triangulation = tortri.calc_triangulation()

    def generate_slice_attime(self, time):
        """ Specialised version of the slice version, interpolates onto the finer z grid"""
        slice3d_t = np.pad(super().generate_slice_attime(time), (0, 1), mode="wrap")
        pos_y = self._get_nearest_y_pos()
        posyrange = np.arange(0, self.cm.pnt.nky0*2 + 1)
        interp_data = np.empty((self.cm.pnt.nx0, self.nz_extended))

        for ix in range(len(self.cm.spatialgrid.x_a)):
            data_interp = interp.RectBivariateSpline(posyrange, self.z_full, slice3d_t[ix, ...])
            interp_data[ix, :] = data_interp(pos_y[ix, :], self.z_full_fine, grid=False)
        return interp_data

    def _get_nearest_y_pos(self):
        """ Calculate which y position we need for the toroidal slice """
        z = self.z_full_fine
        qprof = self.cm.pnt.q0*(
                1 + self.cm.pnt.shat/self.cm.pnt.x0*(self.cm.spatialgrid.x_a - self.cm.pnt.x0))
        # Get nearest y for given torangle \phi, y=C_y*(q(x)z - \phi)
        y_near = self.geom.Cy*(np.outer(qprof, z)*self.cm.pnt.sign_Ip_CW - self.torangle)
        ny = 2*self.cm.pnt.nky0
        dy = self.cm.pnt.rhostar*self.cm.pnt.ly/ny
        pos_y_near = (y_near/dy + ny/2)%ny
        return pos_y_near


class TorusCutTriangulation(object):
    """ Set up the triangulation of a torus cut based on GENE geometry outpout"""

    def __init__(self, common, z_res):
        """
        :param common: CommonData object of the run
        :param z_res: The number of points that z should be interpolated to
        """
        self.cm = common
        self.geom = self.geom = pydiag.utils.geom.Geometry(common)
        self.z_full = np.append(self.cm.spatialgrid.z, np.pi)  # Extend the z grid to pi
        self.z_full_fine = np.linspace(self.z_full[0], self.z_full[-1], z_res)
        self.triangulation = None

    def calc_triangulation(self, recalculate=False):
        """ Calculate the triangulation based on local/global

        :param recalculate: Force recalculation of the triangulation, otherwise a previous result is
        used
        """
        if not self.cm.y_local:
            raise RuntimeError("y-global is not supported in torus cut visualisation")
        if (not self.triangulation) or recalculate:
            if self.cm.x_local:
                return self._calc_triangulation_local()
            else:
                return self._calc_triangulation_xglobal()
        else:
            return self.triangulation

    def _calc_triangulation_xglobal(self):
        raise NotImplementedError("Torus viz for x-global is still work-in-progress")
        return None

    def _calc_triangulation_local(self):
        """ Calculate a triangulation based on R(x,z), Z(x,z) recovered from flux-tube geometry"""
        # These are the values at x=x0 extended so that z covers [-pi, pi]
        Rz = np.pad(self.geom.R, (0, 1), mode="wrap")
        Zz = np.pad(self.geom.Z, (0, 1), mode="wrap")
        # Interpolate for the x direction
        dxdR = np.pad(self.geom.dxdR/self.geom.gxx, (0, 1), mode="wrap")
        dxdZ = np.pad(self.geom.dxdZ/self.geom.gxx, (0, 1), mode="wrap")
        x_phys = (self.cm.spatialgrid.x_a - self.cm.pnt.x0)*self.cm.pnt.Lref
        R_pos = interp.RectBivariateSpline(x_phys, self.z_full, Rz + np.outer(x_phys, dxdR))(x_phys,
                                                                                             self.z_full_fine)
        Z_pos = interp.RectBivariateSpline(x_phys, self.z_full, Zz + np.outer(x_phys, dxdZ))(x_phys,
                                                                                             self.z_full_fine)
        # Calculate a triangulation based on R(x,z) and Z(x,z)
        RZtri = mtri.Triangulation(R_pos.flatten(), Z_pos.flatten())
        # Mask central hole
        centretri = mtri.Triangulation(R_pos[0, :], Z_pos[0, :])
        centermask = np.full(RZtri.triangles.shape[0], False)
        for ct in centretri.triangles:
            start = np.where(ct.mean() == RZtri.triangles.mean(axis=1))
            centermask[start] = True
        RZtri.set_mask(centermask)
        self.triangulation = RZtri
        return self.triangulation


class PlotTorusCut(pydiag.diagplots.baseplot.Plotting):
    """ Class handling the plotting of the triangulated torus cuts"""

    def __init__(self, toruscutlist):
        super().__init__()
        self.toruscutlist = toruscutlist
        self.anims = []  # The animation objects need to be stored for the animations to run
        self.figlist = []
        self.ax = None

    def createfigs(self, animate=False, save=False):
        for torus in self.toruscutlist:
            if not animate:
                torus.generate_timeaverage()
                fig = plt.figure()
                ax = fig.add_subplot(1, 1, 1)
                ax.set_aspect('equal', 'box')
                ax.tricontourf(torus.triangulation, torus.timeaverage.flatten(), 50,
                               cmap=self.cmap_unidirect)
                ax.set_xlabel(r"$R[m]$")
                ax.set_ylabel(r"$Z[m]$")
                fig.tight_layout()
            else:
                torus.generate_timeseries()
                maxval = np.max(torus.dataarray)
                minval = np.min(torus.dataarray)
                self.figlist.append(plt.figure())
                self.ax = self.figlist[-1].add_subplot(111)
                self.ax.set_aspect('equal', 'box')
                self.ax.tricontourf(torus.triangulation, torus.dataarray[0].flatten(), 50,
                                    cmap=self.cmap_unidirect, vmax=maxval, vmin=minval, zorder=-20)

                def on_click(event):
                    if event.dblclick:
                        for ani in self.anims:
                            if ani.running:
                                ani.event_source.stop()
                            else:
                                ani.event_source.start()
                            ani.running ^= True

                def animate(t):
                    self.ax.clear()
                    self.ax.set_xlabel(r"$R[m]$")
                    self.ax.set_ylabel(r"$Z[m]$")
                    self.ax.set_rasterization_zorder(z=-10)
                    self.ax.set_title(r"t={:.3f}".format(torus.timearray[t]))
                    cma = self.ax.tricontourf(torus.triangulation, torus.dataarray[t].flatten(), 50,
                                              cmap=self.cmap_unidirect, vmax=maxval, vmin=minval,
                                              zorder=-20)
                    return cma

                self.anims.append(animation.FuncAnimation(self.figlist[-1], animate,
                                                          range(len(torus.timearray)), blit=False))
                for anim in self.anims:
                    anim.running = True
                for fig in self.figlist:
                    fig.canvas.mpl_connect('button_press_event', on_click)
                if save:
                    # Keep the plotting from freezing by doing the saving in a subprocess
                    def save_anim():
                        print("Saving animation in the background, this can take a long while")
                        self.anims[-1].save(
                                "animatedtorus_{}_{}{}.mp4".format(torus.quantity, torus.species,
                                                                   torus.cm.fileextension), dpi=300)
                        print("Finished saving animation")

                    p = multiprocessing.Process(target=save_anim)
                    p.daemon = True
                    p.start()



"""
Module handling nrg files

"""
import csv
import numpy as np

from .base_file import TimeSeries


class NrgFile(TimeSeries):
    """NrgFile: Class to read a nrg file from GENE

    :param filename: Typically nrg.dat
    :param common: CommonData object of the run
    """
    def __init__(self, filename, common):
        super().__init__(filename, common)
        self.dataarray = []
        self.n_col = common.pnt.nrgcols
        self.n_spec = common.pnt.n_spec

    def generate_timeseries(self):
        """ Fill the Nrgdata object with data """
        n_col = self.n_col
        self.timearray = []
        self.dataarray = []
        try:
            with open(self.objectname) as nrgfile:
                csvnrg = csv.reader(nrgfile, delimiter=' ', skipinitialspace=True)
                for line in csvnrg:
                    if len(line) == 0:
                        continue
                    if len(line) == 1:
                        self.timearray.append(float(line[0]))
                        self.dataarray.append([[] for _ in range(self.n_spec)])
                        ispec = 0  # Reset index for the species at the current time step
                    elif len(line) == n_col:
                        self.dataarray[-1][ispec] = line
                        ispec += 1
                    else:
                        raise IOError("Incorrect number of columns")
        except IOError:
            raise IOError("nrg file does not exist or has"
                          " wrong number of columns: {}".format(self.objectname))
        self.check_times()
        pos = self.calc_positions()
        self.dataarray = np.array(self.dataarray).astype(float, copy=False)
        # Reduce the nrgcols to only the required time frame
        self.dataarray = self.dataarray[pos, ...]
        self.timearray = np.array(self.timearray)[pos]


def gluenrgdata(nrglist):
    """ Function to combine a list of nrg files from a continuation run.

    :param nrglist: the list of NrgdFile objects to combine
    :returns: the combined NrgFile object
    """
    # TODO: More sanity checks for agreeing parameters in the list
    if len(nrglist) == 1:
        return nrglist[0]
    result = nrglist.pop(0)
    for nrg in nrglist:
        # When times overlap, give following Nrgdata preference
        nrg.timearray = nrg.timearray.tolist()
        result.timearray = result.timearray.tolist()
        try:
            while result.timearray[-1] >= nrg.timearray[0]:
                del result.timearray[-1]
        except IndexError:
            raise IndexError("The nrg files completely overlap")
        result.dataarray = result.dataarray[:len(result.timearray), ...]
        result.timearray = np.concatenate((result.timearray, nrg.timearray))
        result.dataarray = np.concatenate((result.dataarray, nrg.dataarray))
    result.endtime = nrglist[-1].endtime
    del nrglist
    return result



# -*- coding: utf-8 -*-
"""
Created on Wed Nov 21 14:20:41 2018

"""
import numpy as np
from scipy.sparse import spdiags


def compute(dx, F, order, accuracy=None):
    F = np.reshape(F, (len(F), 1))

    if order == 1:
        return derivativeO1(dx, F, accuracy if accuracy else 2)
    elif order == 2:
        return derivativeO2(dx, F, accuracy if accuracy else 3)
    else:
        raise NotImplementedError("Not implemented derivatives higher than second order")


def derivativeO1(dx, F, accuracy):
    m = len(F)
    e = np.ones((1, m))
    z = np.zeros((1, m))
    if accuracy == 2:
        D = spdiags(np.row_stack((-0.5*e, z, 0.5*e)), np.arange(-1, 2), m, m)
    elif accuracy == 4:
        D = spdiags(np.row_stack((1/12*e, -2/3*e, z, 2/3*e, -1/12*e)), np.arange(-2, 3), m, m)
    elif accuracy == 6:
        D = spdiags(np.row_stack((-1/60*e, 3/20*e, -3/4*e, z, 3/4*e, -3/20*e, 1/60*e)),
                    np.arange(-3, 4), m, m)
    elif accuracy == 8:
        D = spdiags(
            np.row_stack((1/280*e, -4/105*e, 1/5*e, -4/5*e, z, 4/5*e, -1/5*e, 4/105*e, -1/280*e)),
            np.arange(-4, 5), m, m)
    else:
        raise NotImplementedError("Not implemented accuracy other than 2, 4, 6 or 8")

    dF = D*F
    dF[0] = F[1] - F[0]
    dF[-1] = F[-1] - F[-2]

    return dF/dx


def derivativeO2(dx, F, accuracy):
    m = len(F)
    e = np.ones((m, 1))
    if accuracy != 3:
        raise NotImplementedError("Not implemented accuracy other than 3")

    D = spdiags(np.row_stack((e, -2*e, e)), np.arange(-1, 2), m, m)

    dF = D*F
    dF[0] = dF[1]
    dF[-1] = dF[-2]

    return dF/dx


# -*- coding: utf-8 -*-

import h5py
import matplotlib.pyplot as plt
import numpy as np

import ..utils.averages as averages
from .baseplot import Plotting
from .diagnostic import Diagnostic


class DiagSrc(Diagnostic):
    def __init__(self, avail_vars=None, specnames=None):
        super().__init__()
        self.name = 'Sources'
        self.tabs = ['xglobal', "GENE3D"]

        self.help_txt = """ Analyze sources
        \nt avg : time averaged (def. True)
        \nSave h5 : save hdf5 file (def.True) """

        self.avail_vars = avail_vars
        self.specnames = specnames

        self.opts = {'t_avg': {'tag': "t avg", 'values': [True, False]},
                     "save_h5": {'tag': 'Save h5', 'values': [True, False]}}

        self.set_defaults()

        self.options_gui = Diagnostic.OptionsGUI()

    def set_options(self, run_data, specnames=None):
        class SrcStep:
            def __init__(self, specnames, src_types):
                for spec in specnames:
                    for src_type in src_types:
                        setattr(self, spec + src_type, self.__SrcMoms())

            class __SrcMoms:
                def __init__(self):
                    self.M00 = []
                    self.M10 = []
                    self.M22 = []

        self.run_data = run_data

        self.geom = run_data.geometry

        # select which source is active
        self.src_type = []
        if self.run_data.pnt.ck_part > 0.:
            self.src_type.append('ck_part')
        if self.run_data.pnt.ck_heat > 0.:
            self.src_type.append('ck_heat')
        for spec in self.specnames:
            if "src_amp" + spec in self.run_data.pnt:
                #               I think this is wrong
                if self.run_data.pardict["src_amp" + spec] > 0.:
                    self.src_type.append('f0_term')
                    break

        self.src_step = SrcStep(self.specnames, self.src_type)

        # toggle the reader
        self.get_needed_vars(['srcmom'])

        return self.needed_vars

    def execute(self, data, step):
        for spec in self.specnames:
            for src_type in self.src_type:
                src_step = getattr(self.src_step, spec + src_type)

                src_step.M00.append(getattr(getattr(data, 'srcmom_' + spec), src_type + "_M00")(step.time, step.srcmom))
                src_step.M10.append(getattr(getattr(data, 'srcmom_' + spec), src_type + "_M10")(step.time, step.srcmom))
                src_step.M22.append(getattr(getattr(data, 'srcmom_' + spec), src_type + "_M22")(step.time, step.srcmom))

    def plot(self, time_requested, output=None, out_folder=None):

        def plot_a_map(ax, x, y, f, x_lbl, y_lbl, ttl):
            cm1 = ax.pcolormesh(x, y, f)
            ax.set_rasterization_zorder(z=-10)
            ax.set_xlabel(x_lbl)
            ax.set_ylabel(y_lbl)
            ax.set_title(ttl)
            fig.colorbar(cm1)

        self.plotbase = Plotting()

        y_lbl = r'x/a'
        x_lbl = r'$t\;c_{ref}/L_{ref}$'

        if len(time_requested) > 1:
            n_col = len(self.src_type)
            for spec in self.specnames:
                fig = plt.figure(figsize=(15, 10), dpi=100)
                for isrc, src in enumerate(self.src_type):
                    ax = fig.add_subplot(3, n_col, 1 + isrc)
                    ttl = r"$" + spec + src.replace('_', '_{') + r"}\; M00$"
                    plot_a_map(ax, time_requested, self.run_data.spatialgrid.x_a, np.squeeze(
                        np.array(getattr(getattr(self.src_step, spec + src), 'M00')).T), x_lbl, y_lbl, ttl)
                    ax = fig.add_subplot(3, n_col, 1 + n_col + isrc)
                    ttl = r"$" + spec + src.replace('_', '_{') + r"}\; M10$"
                    plot_a_map(ax, time_requested, self.run_data.spatialgrid.x_a, np.squeeze(
                        np.array(getattr(getattr(self.src_step, spec + src), 'M10')).T), x_lbl, y_lbl, ttl)
                    ax = fig.add_subplot(3, n_col, 1 + 2*n_col + isrc)
                    ttl = r"$" + spec + src.replace('_', '_{') + r"}\; M22$"
                    plot_a_map(ax, time_requested, self.run_data.spatialgrid.x_a, np.squeeze(
                        np.array(getattr(getattr(self.src_step, spec + src), 'M22')).T), x_lbl, y_lbl, ttl)
                fig.show()

        if self.opts['t_avg']['value']:
            # time averaged values
            for spec in self.specnames:
                fig = plt.figure(figsize=(15, 10), dpi=100)
                ax_M00 = fig.add_subplot(3, 1, 1)
                ax_M10 = fig.add_subplot(3, 1, 2)
                ax_M22 = fig.add_subplot(3, 1, 3)
                M00 = np.zeros(self.run_data.spatialgrid.x.shape)
                M10 = np.zeros(self.run_data.spatialgrid.x.shape)
                M22 = np.zeros(self.run_data.spatialgrid.x.shape)
                for isrc, src in enumerate(self.src_type):

                    tmp = averages.mytrapz(getattr(getattr(self.src_step, spec + src), 'M00'),
                                           time_requested)
                    ax_M00.plot(self.run_data.spatialgrid.x_a, tmp, label=src)
                    M00 = M00 + np.squeeze(tmp)

                    tmp = averages.mytrapz(getattr(getattr(self.src_step, spec + src), 'M10'),
                                           time_requested)
                    ax_M10.plot(self.run_data.spatialgrid.x_a, tmp, label=src)
                    M10 = M10 + np.squeeze(tmp)

                    tmp = averages.mytrapz(getattr(getattr(self.src_step, spec + src), 'M22'),
                                           time_requested)
                    ax_M22.plot(self.run_data.spatialgrid.x_a, tmp, label=src)
                    M22 = M22 + np.squeeze(tmp)

                ax_M00.plot(self.run_data.spatialgrid.x_a, M00.T, label='Total')
                ax_M00.set_xlabel(y_lbl)
                ax_M00.set_ylabel("M00")
                ax_M00.set_title(r"${}$".format(spec))
                ax_M00.legend()
                ax_M10.plot(self.run_data.spatialgrid.x_a, M10.T, label='Total')
                ax_M10.set_xlabel(y_lbl)
                ax_M10.set_ylabel("M10")
                ax_M10.legend()
                ax_M22.plot(self.run_data.spatialgrid.x_a, M22.T, label='Total')
                ax_M22.set_xlabel(y_lbl)
                ax_M22.set_ylabel("M22")
                ax_M22.legend()
                fig.show()

                if self.opts['save_h5']['value']:
                    file = h5py.File(out_folder / 'sources_{}.h5'.format(spec), 'w')
                    file["/x_a"] = self.run_data.spatialgrid.x_a
                    file["/M00"] = M00
                    file["/M10"] = M10
                    file["/M22"] = M22
                    file.close()



#!/usr/bin/env python3
import math

class Units:
    '''
    pnt: Parameters object

    '''
    def __init__(self, parameters):
        self.mp = 1.672621637E-27
        self.e = 1.6021766208E-19
        self.c = 2.99792458e10
    
        self.Bref = parameters.pardict["Bref"]  # (in T)
        self.nref = parameters.pardict["nref"]  # (in 10^19)
        self.Tref = parameters.pardict["Tref"]  # (in ev)
        self.Lref = parameters.pardict["Lref"]  # (in m)
        self.mref = parameters.pardict["mref"]  # in proton masses
        self.qref = 1.0
        # Coulomb logarithm
        self.lnlambda = 24. - math.log(math.sqrt(self.nref * 1e13) /\
                                       self.Tref * 1e3)
        # electron beta
        self.beta = 403e-5 * self.nref * self.Tref / self.Bref** 2
        # Collisionlity
        self.coll = 2.3031e-5 * self.nref / self.Tref**2 * self.Lref * \
            self.lnlambda*1e2  # assume Tref=Te

        self.cref = math.sqrt(self.Tref * 1e3 * self.e/ (self.mref * self.mp))
        self.gyrofreq = self.qref * self.e * self.Bref / (self.mref * self.mp)
        self.rhoref = self.cref / self.gyrofreq
        self.inv_rhostar = self.Lref / self.rhoref
        self.pref = self.nref * 1e19 * self.Tref * 1e3 * self.e
        self.GGB = self.nref * 1e19* self.cref * (self.rhoref / self.Lref)**2
        self.QGB = self.pref * self.cref * (self.rhoref / self.Lref)**2
        self.PGB = self.mref *self.mp * self.nref * 1e19 * self.cref**2 * (self.rhoref / self.Lref)**2

        try:
            Zeff_in = parameters.pardict['Zeff']
        except:
            Zeff_in = 1.0

        qi = 1.
        ne = 1.
        te = 1.
        ti = 1.
        self.tau = 1.
        self.mi = 1.
        self.Zeff = 0.
        for spec in parameters.specnames:
            if parameters.pardict['charge' + spec] < 0:  # for electrons
                ne = parameters.pardict['dens' + spec]
                te = parameters.pardict['temp' + spec]
            else:
                if (parameters.pardict['charge' + spec] == 1): # for main ion species
                    ti = parameters.pardict['temp' + spec]
                    qi = parameters.pardict['charge' + spec]
                    mi = parameters.pardict['mass' + spec]
                self.Zeff = self.Zeff + parameters.pardict['charge' + spec]**2 * \
                    parameters.pardict['dens' + spec]      
        self.Zeff = self.Zeff / ne
        
        '''at this point self. Zeff is either the Zeff or 0 if adiabatic ions.
            Fix latter case'''
        if self.Zeff ==0:
            self.Zeff = Zeff_in
        
        self.tau = ti / te


""" This module contains a number of (semi-)analytic predictions from neoclassical theory"""

import numpy as np
from .averages import mytrapz
import scipy.integrate as intg
import scipy.special as sps


def calc_potwidth(prd, ispec=1):
    """ Calculate an estimate for the potato width

    :param prd: ProfileData object of a run
    :param ispec: Index of the species we want, starting from 1
    :returns: potato width for the first species (should be main ions)
    """
    mspec = prd.cm.pars["mass{}".format(ispec)]
    potwidth = (4*prd.q[0]**2 * prd.cm.pnt.rhostar**2 * prd.cm.pnt.major_R * prd.cm.pnt.minor_r**2 *
                2*mspec*prd.T0s[0])**(1.0/3)
    return potwidth


def chang_hinton(prd, species):
    """ Calculate the Chang-Hinton estimate for the neoclassical heat flux

    :param prd: ProfileData object of a run
    :param species: Index of species to use
    :returns: Radial profile (numpy array) with the C-H prediction
    """
    # TODO: Implement the more general case for multiple ion species (Z_eff != 1)
    Qcharr = []
    epsilon = np.array(prd.xs * prd.cm.pnt.minor_r / prd.cm.pnt.major_R)
    mspec = 1  # since we only support one ion species so far
    prspec = prd.prspec[species]
    # alpha = Zeff - 1
    Z = 1
    f1 = 1 + 1.5 * epsilon ** 2
    f2 = (1 - epsilon ** 2) ** 0.5
    try:
        coll = prd.cm.pnt.coll
    except AttributeError:
        coll = 0
    # for alpha == 0: mustar == nustar
    for tb in range(len(prspec.dataarray["Ts"])):
        # k1 = 0 # relevant only for Zeff != 1
        mustar = prspec.nustar[tb]
        T = prspec.dataarray["Ts"][tb]
        n = prspec.dataarray["ns"][tb]
        omt = prspec.dataarray["omts"][tb]
        k2 = ((0.66 + 1.88 * epsilon ** 0.5 - 1.54 * epsilon) /
              (1 + 1.03 * mustar ** 0.5 + 0.31 * mustar) * f1 +
              1.6 * 0.74 * mustar * epsilon /
              (1 + 0.74 * mustar * epsilon ** 1.5) * 0.5 * (f1 - f2))
        qchval = (16.0 / 3 / np.sqrt(np.pi) * n ** 2 *
                  prd.q ** 2 / epsilon ** 1.5 * (1 - epsilon ** 2) *
                  Z ** 4 * coll * np.sqrt(2 * T * mspec) * k2 * omt)
        Qcharr.append(qchval)
    return np.array(Qcharr)


def fb_param_hinton_haz(prd):
    """ Calculate the force balance parameter k based on Hinton, Hazeltine

    :param prd: Profile data object
    :returns: numpy array of k for each radial position
    """
    avnustar = mytrapz(prd.prspec[0].nustar, prd.prspec[0].timearray)
    eps = np.array(prd.xs * prd.cm.pnt.minor_r / prd.cm.pnt.major_R)
    hhpred1 = (1.17 - 0.35 * avnustar ** (1 / 2)) / (1 + 0.7 * avnustar ** (1 / 2) * eps ** 3)
    hhpred2 = (hhpred1 - 2.1 * avnustar ** 2 * eps ** 3) / (1 + avnustar ** 2 * eps ** 3)
    return hhpred2


def fb_param_hirsch_sig(prd):
    """ Calculate the force balance parameter k based on Hirschman, Sigmar, 1981 and Satake 2010
    :param prd: Profile data object
    :returns: numpy array of k for each radial position
    """
    eps = np.array(prd.xs * prd.cm.pnt.minor_r / prd.cm.pnt.major_R)
    ft = 1.96*np.sqrt(eps)
    fc = 1-ft
    nustar = mytrapz(prd.prspec[0].nustar, prd.prspec[0].timearray)
    k = np.zeros(prd.cm.pnt.nx0)
    for i in range(prd.cm.pnt.nx0):
        K11, eK11 = intg.quad(unintK, 0, 5, args=(1, 1, nustar[i], eps[i]))
        K12, eK12 = intg.quad(unintK, 0, 5, args=(1, 2, nustar[i], eps[i]))
        K13, eK13 = intg.quad(unintK, 0, 5, args=(1, 3, nustar[i], eps[i]))
        mu1 = K11
        mu2 = K12 - 2.5*K11
        mu3 = K13 - 5*K12 + 25.0/4*K11
        k[i] = -np.sqrt(2)*mu2/(np.sqrt(2)*mu1 + ft[i]/fc[i]*(mu1*mu3-mu2**2))
    return np.array(k)


def unintK(x, i, j, nustar, eps):
    """ The matrix elements K_ij before x integration """
    nustar = np.array(nustar)
    return 8/3/np.sqrt(np.pi)*x**4*np.exp(-x**2)*x**(2*(i+j-2))*nutottau(x, nustar, eps)


def nutottau(x, nustar, eps):
    """ nu_tot(x)*tau_ii"""
    c = 3*np.sqrt(np.pi)/4
    return c*nu_d_nuhat(x)/(1+nustar*nu_d_nuhat(x)*c/x)/(1+15*np.pi**(3/2)/32*nustar*eps*nu_t_omT(x)/x)


def nu_d_nuhat(x):
    return (sps.erf(x)-chandr(x))/x**3


def nu_t_omT(x):
    return (sps.erf(x)-3*chandr(x))/x**3 + 8*chandr(x)/x


def chandr(x):
    """ The Chandrasehkar function"""
    return (sps.erf(x) - x*2/np.sqrt(np.pi)*np.exp(-x**2))/2/x**2



import numpy as np


def mytrapz(yvar, timefld):
    """ Trapezoid rule, adjusted for single timestep

    Operates on the first dimension of yvar (typically time)
    hence the name timefld for the integration variable samples
    :param yvar: Variable to integrate over
    :param timefld: The integration variable (time)
    :returns: The integration result (dim(yvar)-1)
    """
    timefld = np.array(timefld)
    yvar = np.real_if_close(yvar)
    if timefld.size == 1:
        return np.atleast_2d(yvar)[0]
    else:
        if yvar.shape[0] != len(timefld):
            raise ValueError("First dimension of yvar and timefld do not match")
        tmpt = timefld/(timefld[-1] - timefld[0])
        return np.trapz(yvar, x=tmpt, axis=0)


def z_av3d(var, geom):
    """ Perform the average in z direction for a 3d variable

    :param var: Variable to average over
    :param geom: GENE geomtry object

    """
    if geom.cm.x_local and geom.cm.y_local:
        return np.average(var, weights=geom.jacobian, axis=-1)
    else:
        return np.average(var, weights=geom.jaco3d.T, axis=-1)  # geom has (z, x or y) arrays


def y_av3d(var, geom):
    """ Perform the average in y direction for a 3d variable

    :param var: Variable to average over
    :param geom: GENE geomtry object
    """
    if geom.cm.y_local:  # Add the negative half of the Fourier space for ky
        var[:, 0, :] *= 0.5
        return np.sum(2*var, axis=1)
    else:
        return np.average(var, weights=geom.jaco3d.T, axis=1)


def x_av3d(var, geom):
    """ Perform the average in x direction for a 3d variable

    :param var: Variable to average over
    :param geom: GENE geomtry object
    """
    if geom.cm.x_local:
        if geom.cm.y_local:
            return np.sum(var, axis=0)
        else:  # Mirror positive kx to negative kx
            var[0, :, :] *= 0.5
            return np.sum(2*var, axis=0)
    else:
        return np.average(var, weights=geom.jaco3d.T, axis=0)


def xz_av3d(var, geom):
    """ Perform the average in x and z direction for a 3d variable

    :param var: Variable to average over
    :param geom: GENE geomtry object
    """
    if geom.cm.x_local:
        if geom.cm.y_local:
            return np.average(np.sum(var, axis=0), weights=geom.jacobian, axis=-1)
        else:  # Mirror positive kx to negative kx
            var[0, :, :] *= 0.5
            return np.average(np.sum(2*var, axis=0), weights=geom.jacobian.T, axis=-1)
    else:
        return np.average(var, weights=geom.jaco3d.T, axis=(0, -1))


def flux_spectra_xz_av(var, geom):
    """ Perform xz average in the form required for ky flux spectra (see diagplots.plot_spectra)"""
    if geom.cm.x_local:
        if geom.cm.y_local:
            nx0o2 = int(geom.cm.pnt.nx0/2)
            temp = np.empty((geom.cm.pnt.nky0, geom.cm.pnt.nz0), dtype=np.complex128)
            temp[0, :] = var[0, 0, :] + var[nx0o2, 0, :] + np.sum(var[1:nx0o2, 0, :], axis=0)
            temp[1:, :] = 2*np.real(np.sum(var[:, 1:, :], axis=0))
            return np.average(temp, weights=geom.jacobian, axis=-1)
        else:
            raise NotImplementedError("No support for y-global")
    else:
        raise NotImplementedError("No support for x-global yet")


def yz_av3d(var, geom):
    """ Perform the average in y and z direction for a 3d variable

    :param var: Variable to average over
    :param geom: GENE geomtry object
    """

    if geom.cm.y_local:  # Add the negative half of the Fourier space for ky
        # The transposition of the Jacobian only matters for x-global
        return np.average(np.sum(2*var, axis=1)-var[:, 0, :], weights=geom.jacobian.T, axis=-1)
    else:
        return np.average(var, weights=geom.jaco3d.T, axis=(1, -1))  # geom has (z, x or y) arrays


def flux_spectra_yz_av(var, geom):
    """ Perform yz average in the form required for kx flux spectra (see diagplots.plot_spectra)"""
    if not geom.cm.y_local:
        raise NotImplementedError("No support for y-global")
    if geom.cm.x_local:
        nx0o2 = int(geom.cm.pnt.nx0/2)
        temp = np.zeros((nx0o2+1, geom.cm.pnt.nz0))
        temp[0, :] = np.real_if_close(var[0, 0, :]) + 2*np.sum(np.real(var[0, 1:, :]), axis=0)

        temp[1:nx0o2, :] = 2*np.real_if_close(var[1:nx0o2, 0, :]) + 2*np.sum(np.real(
                var[1:nx0o2, 1:, :] + var[2*nx0o2:nx0o2:-1, 1:, :]), axis=1)
        temp[nx0o2, :] = np.real_if_close(var[nx0o2, 0, :]) + 2*np.sum(np.real(var[nx0o2, 1:, :]), axis=0)
        return np.average(temp, weights=geom.jacobian, axis=-1)
    else:
        raise NotImplementedError("No support for x-global yet")


def xy_av3d(var, geom):
    """ Perform the average in x and y direction for a 3d variable

    :param var: Variable to average over
    :param geom: GENE geomtry object
    """
    if geom.cm.x_local:
        if geom.cm.y_local:
            var[:, 0, :] *= 0.5
            return np.sum(2*var, axis=(0, 1))
        else:  # Mirror positive kx to negative kx
            var[0, :, :] *= 0.5
            return np.average(np.sum(2*var, axis=0), weights=geom.jacobian.T, axis=1)
    else:
        if geom.cm.y_local:
            var[:, 0, :] *= 0.5
            return np.average(np.sum(2*var, axis=1), weights=geom.jacobian.T, axis=0)
        else:
            return np.average(var, weights=geom.jaco3d.T, axis=(0, 1))


def xyz_av3d(var, geom):
    """ Perform the average in all spatial directions for a 3d variable

    :param var: Variable to average over
    :param geom: GENE geomtry object
    """
    if geom.cm.x_local:
        if geom.cm.y_local:
            var[:, 0, :] *= 0.5
            return np.average(np.sum(2*var, axis=(0, 1)), weights=geom.jacobian)
        else:  # Mirror positive kx to negative kx
            var[0, :, :] *= 0.5
            temp = np.sum(2*var, axis=0)
            return np.average(temp, weights=geom.jacobian.T, axis=(0, -1))
    else:
        if geom.cm.y_local:
            var[:, 0, :] *= 0.5
            return np.average(np.sum(2*var, axis=1), weights=geom.jacobian.T, axis=(0, -1))
        else:
            return np.average(var, weights=geom.jaco3d.T)


def av3d_by_switch(xavg, yavg, zavg):
    """ Map the required averaging function by binary switches for the averages

    :returns: Reference to the appropriate function from this module
    """
    if not (xavg or yavg or zavg):
        return lambda var, geom: var  # Return an identity function
    if xavg and not (yavg or zavg):
        return x_av3d
    if xavg and yavg and not zavg:
        return xy_av3d
    if xavg and not yavg and zavg:
        return xz_av3d
    if xavg and yavg and zavg:
        return xyz_av3d
    if not (xavg or yavg) and zavg:
        return z_av3d
    if not xavg and yavg and zavg:
        return yz_av3d
    if not xavg and yavg and not zavg:
        return y_av3d



# -*- coding: utf-8 -*-

from tkinter import END
import h5py
import matplotlib.pyplot as plt
import numpy as np
import ..utils.aux_func as aux_func
import ..utils.averages as averages
from .baseplot import Plotting
from .diagnostic import Diagnostic


class DiagAmplitudeSpectragene3d(Diagnostic):

    def __init__(self, avail_vars=None, specnames=None):
        super().__init__()
        self.name = 'Amplitude spectra'
        self.tabs = ['GENE3D']

        self.help_txt = """Plot amplitude spectra in ky space,
                        \naveraged over the remaining coordinates and times.
                        \n
                        \nQuantity : quantity to plot
                        \nspec: which species to plot (def. all if spec dependent)
                        \nx ind : comma separated radial averaged interval in x/a (e.g. 0.4,0.6),
                        \nz ind : comma separated z interval (e.g., -3,3), default all range
                        default all range
                        \nSave h5 : save hdf5 file (def. True)"""

        self.avail_vars = avail_vars
        self.specnames = specnames

        self.opts = {"quant": {'tag': "Quant", 'values': None, 'files': ['mom', 'field']},
                     "spec": {'tag': "Spec", 'values': self.list_specnames()},
                     "xind": {'tag': "x ind", 'values': None},
                     "zind": {'tag': "z ind", 'values': None},
                     'save_h5': {'tag': "Save h5", 'values': [True, False]}, }

        self.set_defaults()

        self.options_gui = Diagnostic.OptionsGUI()

    def set_options(self, run_data, specnames):

        if self.opts['quant']['value'] is None:
            raise RuntimeError("No quantities given for contours")

        # radial position, unset or -1 means all
        xind = self.opts['xind']['value']
        if not xind or xind == -1:
            self.xind = 0, 100  # this will choose the minimum and maximum values
        else:
            self.xind = xind.split(',')
        self.xind_min, self.xind_max = aux_func.find_nearest_points(run_data.spatialgrid.x_a,
                                                                    float(self.xind[0]),
                                                                    float(self.xind[1]))

        # z position, unset or -1 means all
        zind = self.opts['zind']['value']
        if not zind or zind == -1:
            self.zind = -100, 100  # this will choose the minimum and maximum values
        else:
            self.zind = zind.split(',')
        self.zind_min, self.zind_max = aux_func.find_nearest_points(run_data.spatialgrid.z,
                                                                    float(self.zind[0]),
                                                                    float(self.zind[1]))

        self.geom = run_data.geometry
        self.run_data = run_data

        self.myquant = self.opts['quant']['value']

        # This is special since we dont save stuff by default, so just do what we need
        self.get_spec_from_opts()

        # toggle the reader
        self.get_needed_vars()
        self.amplitude_ky = {}

        for file in self.needed_vars:
            # loop over quaitites in that file
            for quant in self.needed_vars[file]:
                if self.needed_vars[file][quant]:
                    if file == 'field':
                        self.amplitude_ky[quant] = []
                    else:
                        for spec in self.specnames:
                            self.amplitude_ky[quant + '#' + spec] = []

        return self.needed_vars

    def execute(self, data, step):

        for file in self.needed_vars.keys():
            # loop over quaitites in that file
            for quant in self.needed_vars[file].keys():
                # loop over quaitites in that file
                if self.needed_vars[file][quant]:

                    if file == 'field':
                        # no species dependency
                        data_in = getattr(getattr(data, file), quant)(step.time,
                                                                      getattr(step, file))

                        temp = np.absolute(np.fft.fft(data_in, axis=1)/self.run_data.pnt.ny0) ** 2
                        jac_xz = np.mean(self.geom.jacobian.T, axis=1)
                        jaco3d_xz = np.broadcast_to(jac_xz[:, np.newaxis, :], (
                        self.run_data.pnt.nx0, self.run_data.pnt.ny0, self.run_data.pnt.nz0))

                        data_ky = np.average(temp[self.xind_min:self.xind_max + 1, :,
                                             self.zind_min:self.zind_max + 1],
                                             weights=jaco3d_xz[self.xind_min:self.xind_max + 1, :,
                                                     self.zind_min:self.zind_max + 1], axis=(0, 2))

                        self.amplitude_ky[quant].append(data_ky)

                    else:
                        # spec dependent
                        for spec in self.specnames:

                            data_in = getattr(getattr(data, file + '_' + spec), quant)(step.time,
                                                                                       getattr(step,
                                                                                               file))

                            temp = np.absolute(
                                np.fft.fft(data_in, axis=1)/self.run_data.pnt.ny0) ** 2
                            jac_xz = np.mean(self.geom.jacobian.T, axis=1)
                            jaco3d_xz = np.broadcast_to(jac_xz[:, np.newaxis, :], (
                            self.run_data.pnt.nx0, self.run_data.pnt.ny0, self.run_data.pnt.nz0))

                            data_ky = np.average(temp[self.xind_min:self.xind_max + 1, :,
                                                 self.zind_min:self.zind_max + 1],
                                                 weights=jaco3d_xz[self.xind_min:self.xind_max + 1,
                                                         :, self.zind_min:self.zind_max + 1],
                                                 axis=(0, 2))

                            self.amplitude_ky[quant + '#' + spec].append(data_ky)

    def plot(self, time_requested, output=None, out_folder=None):

        if output:
            output.info_txt.insert(END, "Amplitude specta:\n")

        self.plotbase = Plotting()
        self.plotbase.titles.update(
                {"phi": r"$\phi$", "n": r"$n$", "u_par": r"u_{//}", "T_par": r"T_{//}",
                 "T_perp": r"T_{\perp}", "Q_es": r"$Q_{es}$", "Q_em": r"$Q_{em}$",
                 "Gamma_es": r"$\Gamma_{es}$", "Gamma_em": r"$\Gamma_{em}$"})

        ky = self.run_data.spatialgrid.ky
        ny0_kymax = int(self.run_data.pnt.ny0/2)

        text_x = 'Avg. over x = [' + '{:.1f}'.format(
                self.run_data.spatialgrid.x_a[self.xind_min]) + ',' + '{:.1f}'.format(
                self.run_data.spatialgrid.x_a[self.xind_max - 1]) + ']  '

        for quant in self.amplitude_ky:

            ind_str = quant.find('#')
            if ind_str == -1:
                ttl = self.plotbase.titles[quant]
            else:
                ttl = self.plotbase.titles[quant[0:ind_str]] + " " + quant[ind_str + 1:]

            amplitude_ky = averages.mytrapz(self.amplitude_ky[quant], time_requested)

            fig = plt.figure(figsize=(15, 10), dpi=100)

            ax_1 = fig.add_subplot(1, 3, 1)
            ax_1.loglog(ky[0:ny0_kymax], amplitude_ky[0:ny0_kymax])
            ax_1.set_xlabel(r"$k_y$")
            ax_1.set_ylabel(ttl)

            ax_2 = fig.add_subplot(1, 3, 2)
            ax_2.plot(ky[0:ny0_kymax], amplitude_ky[0:ny0_kymax])
            ax_2.set_xlabel(r"$k_y$")
            ax_2.set_ylabel(ttl)

            ax_3 = fig.add_subplot(1, 3, 3)
            ax_3.semilogx(ky[0:ny0_kymax], ky[0:ny0_kymax]*amplitude_ky[0:ny0_kymax])
            ax_3.set_xlabel(r"$k_y$")
            ax_3.set_ylabel(r"$k_y$" + ttl)

            fig.suptitle(text_x)
            fig.tight_layout()
            fig.show()

            if self.opts['save_h5']['value']:
                file = h5py.File(out_folder/'amplitude_spectra_{}.h5'.format(quant), 'w')
                file["/ky"] = ky
                file["/" + self.myquant] = self.amplitude_ky[quant]
                file.close()



from tkinter import END

import os
import h5py
import matplotlib.pyplot as plt
import numpy as np
from ..putils.averages import flux_spectra_yz_av, flux_spectra_xz_av, xy_av3d_zprofile, mytrapz
from .baseplot import Plotting
from .diagnostic import Diagnostic
from copy import deepcopy


class DiagFluxSpectra(Diagnostic):
    # pylint: disable=invalid-name
    def __init__(self, avail_vars=None, specnames=None):
        super().__init__()
        self.name = 'Flux spectra'
        self.tabs = ['fluxtube']

    def get_info(self):
        self.need_file={'field': True,
                        'mom': True}

        return self.need_file
    
    def setup_options(self, run):
        self.has_EM = run.parameters.pnt.n_fields>1
        self.specnames = run.parameters.pnt.specnames
        self.kx=run.spatialgrid.kx_pos
        self.ky=run.spatialgrid.ky
        self.z=run.spatialgrid.z
        self.flux_spectra_kx=[]
        self.flux_spectra_ky=[]
        self.flux_profile_z=[]
        self.time = []

    def execute(self, data, run, steps, extensions,time_point):
        #We use species, quantity kx y
        print("Executing flux spectra.")
        __spectra_kx=np.empty((len(self.specnames), 6 if self.has_EM else 3,
                               run.spatialgrid.kx_pos.size))
        __spectra_ky=np.empty((len(self.specnames), 6 if self.has_EM else 3,
                               run.spatialgrid.ky.size))
        __profile_z=np.empty((len(self.specnames), 6 if self.has_EM else 3,
                               run.spatialgrid.z.size))

        self.time.append(time_point)
        vE_x = -1.0j*run.spatialgrid.ky[np.newaxis, :, np.newaxis] \
            * data.field.phi(step=steps['field'], extension=extensions['field'])

        if run.parameters.pnt.n_fields>1:
            B_x = 1.0j*run.spatialgrid.ky[np.newaxis, :, np.newaxis] \
                * data.field.A_par(step=steps['field'], extension=extensions['field'])

            
        #print("56")
        for i_spec,spec in enumerate(run.parameters.specnames):
            n0 = run.parameters.pardict["dens{}".format(spec)]
            T0 = run.parameters.pardict["temp{}".format(spec)]
            mass = run.parameters.pardict["mass{}".format(spec)]
            charge = run.parameters.pardict["charge{}".format(spec)]

            dens = data.mom[i_spec].dens(step=steps['mom'], extension=extensions['mom'])
            T_par = data.mom[i_spec].T_par(step=steps['mom'], extension=extensions['mom'])
            T_perp = data.mom[i_spec].T_perp(step=steps['mom'], extension=extensions['mom'])
            u_par = data.mom[i_spec].u_par(step=steps['mom'], extension=extensions['mom'])

            G_es = vE_x*np.conj(dens)*n0/run.geometry.Cxy
            __spectra_kx[i_spec,0,:]=flux_spectra_yz_av(G_es, run.geometry)
            __spectra_ky[i_spec,0,:]=flux_spectra_xz_av(G_es, run.geometry)
            __profile_z[i_spec,0,:] = xy_av3d_zprofile(np.real(G_es),run.geometry)
            
            Q_es = (vE_x*np.conj(0.5*T_par + T_perp + 3./2.*dens)*n0*T0)/run.geometry.Cxy
            __spectra_kx[i_spec,1,:]=flux_spectra_yz_av(Q_es, run.geometry)
            __spectra_ky[i_spec,1,:]=flux_spectra_xz_av(Q_es, run.geometry)
            __profile_z[i_spec,1,:] = xy_av3d_zprofile(np.real(Q_es),run.geometry)
            #plt.plot(self.__spectra_ky[i_spec,1,:])
            #plt.title(str(time_point))
            #plt.show()
            #plt.plot(self.__spectra_kx[i_spec,1,:])
            #plt.title(str(time_point))
            #plt.show()
            
            P_es = (vE_x*np.conj(u_par)*n0*mass)/run.geometry.Cxy
            __spectra_kx[i_spec,2,:]=flux_spectra_yz_av(P_es, run.geometry)
            __spectra_ky[i_spec,2,:]=flux_spectra_xz_av(P_es, run.geometry)
            __profile_z[i_spec,2,:] = xy_av3d_zprofile(np.real(P_es),run.geometry)
            
            #print("89")
            if run.parameters.pnt.n_fields>1:
                Q_par = data.mom[i_spec].q_par(step=steps['mom'], extension=extensions['mom'])
                Q_perp = data.mom[i_spec].q_perp(step=steps['mom'], extension=extensions['mom'])

                G_em = B_x*np.conj(u_par)*n0/run.geometry.Cxy
                Q_em = B_x*np.conj(Q_par + Q_perp)*n0*T0/run.geometry.Cxy
                P_em = B_x*np.conj(T_par + dens)*n0*mass/run.geometry.Cxy
#TODO verify PEM something is wrong in either GENE or here
                if run.parameters.pnt.bpar:
                    # todo check if the normalization is correct
                    dBpar_dy_norm = -1j*run.spatialgrid.ky[np.newaxis,:,np.newaxis]* \
                            data.field.B_par(step=steps['field'], extension=extensions['field'])/ \
                            run.geometry.Bfield[np.newaxis, np.newaxis,:]*T0/charge
                            
                    densI1 = data.mom[i_spec].densI1(step=steps['mom'], extension=extensions['mom'])
                    TparI1 = data.mom[i_spec].TparI1(step=steps['mom'], extension=extensions['mom'])
                    TppI1 = data.mom[i_spec].TppI1(step=steps['mom'], extension=extensions['mom'])

                    G_em = G_em + (dBpar_dy_norm*np.conj(densI1))*n0/run.geometry.Cxy
                    Q_em = Q_em + (dBpar_dy_norm*np.conj(TparI1 + TppI1))*n0*T0/run.geometry.Cxy
                    
                __spectra_kx[i_spec,3,:]=flux_spectra_yz_av(G_em, run.geometry)
                __spectra_ky[i_spec,3,:]=flux_spectra_xz_av(G_em, run.geometry)
                __profile_z[i_spec,3,:] = xy_av3d_zprofile(np.real(G_em),run.geometry)
                __spectra_kx[i_spec,4,:]=flux_spectra_yz_av(Q_em, run.geometry)
                __spectra_ky[i_spec,4,:]=flux_spectra_xz_av(Q_em, run.geometry)
                __profile_z[i_spec,4,:] = xy_av3d_zprofile(np.real(Q_em),run.geometry)
                __spectra_kx[i_spec,5,:]=flux_spectra_yz_av(P_em, run.geometry)
                __spectra_ky[i_spec,5,:]=flux_spectra_xz_av(P_em, run.geometry)
                __profile_z[i_spec,5,:] = xy_av3d_zprofile(np.real(P_em),run.geometry)
                           
        #plt.plot(self.ky,__spectra_ky[1,1,:])
        #plt.title(str(time_point))
        #plt.show()
        #print("124")
        #if len(self.flux_spectra_ky) > 0:
        #    plt.plot(self.flux_spectra_ky[0][1,1,:])
        #    plt.show()
        self.flux_spectra_kx.append((__spectra_kx))
        self.flux_spectra_ky.append((__spectra_ky))
        self.flux_profile_z.append(__profile_z)


    def dict_to_mgkdb(self):
        self.flux = {}
        self.flux['Description'] = 'Spectra of all fluxes as a function of kx or ky (averaged over other variables).  The structure is [species][which_flux][kx or ky][time list][data on the kx or ky grid]. The z profiles of fluxes (averaged over kx, ky and with a factor of 1/gxx**0.5) are in [species][which flux][z][time list][data on z grid]. The time points are in [time].  The appropriate grids are in [kxgrid], [kygrid], and [zgrid].'  
        #plt.plot(self.flux_spectra_kx[0][0,0,:])
        #plt.show()
        self.flux['kxgrid'] = self.kx
        self.flux['kygrid'] = self.ky
        self.flux['zgrid'] = self.z
        self.flux['time'] = self.time
        for i in range(len(self.specnames)):
            self.flux[self.specnames[i]]={}
            self.flux[self.specnames[i]]['Qes']={}
            self.flux[self.specnames[i]]['Pes']={}
            self.flux[self.specnames[i]]['Ges']={}
            self.flux[self.specnames[i]]['Ges']['kx'] = []
            self.flux[self.specnames[i]]['Qes']['kx'] = []
            self.flux[self.specnames[i]]['Pes']['kx'] = []
            self.flux[self.specnames[i]]['Ges']['ky'] = []
            self.flux[self.specnames[i]]['Qes']['ky'] = []
            self.flux[self.specnames[i]]['Pes']['ky'] = []
            self.flux[self.specnames[i]]['Ges']['z'] = []
            self.flux[self.specnames[i]]['Qes']['z'] = []
            self.flux[self.specnames[i]]['Pes']['z'] = []
                    
            #print('dfs 120')
            if self.has_EM:
                #print('dfs 122')
                self.flux[self.specnames[i]]['Gem']={}
                self.flux[self.specnames[i]]['Qem']={}
                self.flux[self.specnames[i]]['Pem']={}
                self.flux[self.specnames[i]]['Gem']['kx'] = []
                self.flux[self.specnames[i]]['Qem']['kx'] = []
                self.flux[self.specnames[i]]['Pem']['kx'] = []
                self.flux[self.specnames[i]]['Gem']['ky'] = []
                self.flux[self.specnames[i]]['Qem']['ky'] = []
                self.flux[self.specnames[i]]['Pem']['ky'] = []
                self.flux[self.specnames[i]]['Gem']['z'] = []
                self.flux[self.specnames[i]]['Qem']['z'] = []
                self.flux[self.specnames[i]]['Pem']['z'] = []
            for j in range(len(self.flux_spectra_kx)):
                #print('len(self.flux_spectra_kx)',len(self.flux_spectra_kx))
                #print('dfs 133')
                self.flux[self.specnames[i]]['Ges']['kx'].append( self.flux_spectra_kx[j][i,0,:])
                self.flux[self.specnames[i]]['Qes']['kx'].append( self.flux_spectra_kx[j][i,1,:])
                self.flux[self.specnames[i]]['Pes']['kx'].append( self.flux_spectra_kx[j][i,2,:])
                self.flux[self.specnames[i]]['Ges']['ky'].append( self.flux_spectra_ky[j][i,0,:])
                self.flux[self.specnames[i]]['Qes']['ky'].append( self.flux_spectra_ky[j][i,1,:])
                self.flux[self.specnames[i]]['Pes']['ky'].append( self.flux_spectra_ky[j][i,2,:])
                self.flux[self.specnames[i]]['Ges']['z'].append(self.flux_profile_z[j][i,0,:])
                self.flux[self.specnames[i]]['Qes']['z'].append(self.flux_profile_z[j][i,1,:])
                self.flux[self.specnames[i]]['Pes']['z'].append(self.flux_profile_z[j][i,2,:])
                if self.has_EM:
                    self.flux[self.specnames[i]]['Gem']['kx'].append(self.flux_spectra_kx[j][i,3,:])
                    self.flux[self.specnames[i]]['Qem']['kx'].append(self.flux_spectra_kx[j][i,4,:])
                    self.flux[self.specnames[i]]['Pem']['kx'].append(self.flux_spectra_kx[j][i,5,:])
                    self.flux[self.specnames[i]]['Gem']['ky'].append(self.flux_spectra_ky[j][i,3,:])
                    self.flux[self.specnames[i]]['Qem']['ky'].append(self.flux_spectra_ky[j][i,4,:])
                    self.flux[self.specnames[i]]['Pem']['ky'].append(self.flux_spectra_ky[j][i,5,:])
                    self.flux[self.specnames[i]]['Gem']['z'].append(self.flux_profile_z[j][i,3,:])
                    self.flux[self.specnames[i]]['Qem']['z'].append(self.flux_profile_z[j][i,4,:])
                    self.flux[self.specnames[i]]['Pem']['z'].append(self.flux_profile_z[j][i,5,:])
        return self.flux
        
    def plot(self, time_requested, output=None, out_folder=None):
        """ For each selected species we have one figure with six subplots.
            Left is vs. kx, right vs. ky; columnwise we plot, log-log, log-lin, lin-in
            Dashed lines are negative values in log-log plot
            Dashed lines are k multiplied values in log-lin plot"""

        if output:
            output.info_txt.insert(END, "Flux spectra:\n")

        self.plotbase = Plotting()
        self.plotbase.titles={"Ges": r"$\Gamma_{es}$",
                              "Qes": r"$Q_{es}$", 
                              "Pes": r"$\Pi_{es}$"}
        if self.has_EM:
                 self.plotbase.titles.update({"Gem": r"$\Gamma_{em}$", 
                                              "Qem": r"$Q_{em}$", 
                                              "Pem": r"$\Pi_{em}$"})


        for i_spec,spec in enumerate(self.specnames):
            fig = plt.figure(figsize=(6, 8))

            ax_loglog_kx = fig.add_subplot(3, 2, 1)
            ax_loglin_kx = fig.add_subplot(3, 2, 3)
            ax_linlin_kx = fig.add_subplot(3, 2, 5)
            ax_loglog_ky = fig.add_subplot(3, 2, 2)
            ax_loglin_ky = fig.add_subplot(3, 2, 4)
            ax_linlin_ky = fig.add_subplot(3, 2, 6)
            
            for i_flux,flux in enumerate(self.plotbase.titles.keys()):
                if time_requested.size>1:
                    flux_kx = mytrapz(np.asarray(self.flux_spectra_kx)[:,i_spec,i_flux,:], time_requested)
                    flux_ky = mytrapz(np.asarray(self.flux_spectra_ky)[:,i_spec,i_flux,:], time_requested)
                else:
                    flux_kx = np.asarray(self.flux_spectra_kx)[0,i_spec,i_flux,:]
                    flux_ky = np.asarray(self.flux_spectra_ky)[0,i_spec,i_flux,:]
                
                # Mask negative flux values for solid lines
                pos_flux_kx = np.ma.masked_where((flux_kx <= 0), flux_kx)
                # Mask zero flux for dashed lines, this takes care of the Nyquist mode in kx
                all_flux_kx = np.ma.masked_where((flux_kx == 0), flux_kx)

                pos_flux_ky = np.ma.masked_where((flux_ky <= 0), flux_ky)
                all_flux_ky = np.ma.masked_where((flux_ky == 0), flux_ky)

                # log-log =plots, dashed lines for negative values
                baselogkx, = ax_loglog_kx.plot(self.kx, pos_flux_kx, label=self.plotbase.titles[flux])
                ax_loglog_kx.plot(self.kx, np.abs(all_flux_kx), ls="--", color=baselogkx.get_color())
                baselogky, = ax_loglog_ky.plot(self.ky, pos_flux_ky, label=self.plotbase.titles[flux])
                ax_loglog_ky.plot(self.ky, np.abs(all_flux_ky), ls="--", color=baselogky.get_color())

                # lin-log plots, nothing fancy
                baseloglinkx, = ax_loglin_kx.plot(self.kx, all_flux_kx, label=self.plotbase.titles[flux])
                ax_loglin_kx.plot(self.kx, all_flux_kx*self.kx, ls="--", color=baseloglinkx.get_color())
                baseloglinky, = ax_loglin_ky.plot(self.ky, all_flux_ky, label=self.plotbase.titles[flux])
                ax_loglin_ky.plot(self.ky, all_flux_ky*self.ky, ls="--", color=baseloglinky.get_color())

                # lin-lin plots, nothing fancy
                ax_linlin_kx.plot(self.kx, all_flux_kx, label=self.plotbase.titles[flux])
                ax_linlin_ky.plot(self.ky, all_flux_ky, label=self.plotbase.titles[flux])

                str_out = "{} {} = {:.14f} (kx intg.) - {:.14f} (ky intg.)".format(spec, flux,
                                                                                 np.sum(flux_kx),
                                                                                 np.sum(flux_ky))
                if output:
                    output.info_txt.insert(END, str_out + "\n")
                    output.info_txt.see(END)
                else:
                    print(str_out)

            # set things
            ax_loglog_kx.loglog()
            ax_loglog_kx.set_xlabel(r"$k_x \rho_{ref}$")

            ax_loglog_ky.loglog()
            ax_loglog_ky.set_xlabel(r"$k_y \rho_{ref}$")

            ax_loglin_kx.set_xscale("log")
            ax_loglin_ky.set_xscale("log")

            # lin-lin plots, nothing fancy
            ax_linlin_kx.set_xlim(left=0)
            ax_linlin_kx.set_xlabel(r"$k_x \rho_{ref}$")

            ax_linlin_ky.set_xlabel(r"$k_y \rho_{ref}$")

            for ax in [ax_loglog_kx, ax_loglin_kx, ax_linlin_kx, ax_loglog_ky, ax_loglin_ky,
                       ax_linlin_ky, ]:
                # ax.set_ylabel(r"$<|A|^2>$")
                ax.legend()
            ax_loglog_ky.set_title("{}".format(spec))
            ax_loglog_kx.set_title("{}".format(spec))
            #            fig.tight_layout()
            fig.show()

    def save(time_requested, output=None, out_folder=None):
        pass



#!/usr/bin/env python3
""" Command line tool for the python slice diagnostic"""
import argparse
import os

import ..data.slices as sli
import ..utils.comm as comm

xrange = (None, None)
yrange = (None, None)
zrange = (None, None)

parser = argparse.ArgumentParser()
parser.add_argument("time", help="Time window for averaging in the form start:end. "
                                 "Use f for the first and l for the last time in the data file.")
parser.add_argument("fileextension", nargs='+', help="List of file extensions of the runs"
                                                     " to be processed.")
parser.add_argument("--xavg", help="Average over full x direction", action="store_true")
parser.add_argument("--yavg", help="Average over full x direction", action="store_true")
parser.add_argument("--zavg", help="Average over full x direction", action="store_true")
parser.add_argument("--runpath", default=".", help="Path of the run")
parser.add_argument("--specname", help="Name of the species to process")
parser.add_argument("--var", default="phi",
                    help="Name of the variable, any of phi, apar, bpar, dens, tpar, tperp, qpar, "
                         "qperp, upar")

args = parser.parse_args()
os.chdir(args.runpath)

fileextension = args.fileextension[0]

if args.time.split(':')[0] == 'f':
    starttime = -1
elif args.time.split(':')[0] == 'l':
    starttime = -2
else:
    starttime = int(float(args.time.split(':')[0]))
if args.time.split(':')[1] == 'f':
    endtime = -1
elif args.time.split(':')[1] == 'l':
    endtime = -2
else:
    endtime = int(float(args.time.split(':')[1]))

common = comm.CommonData(fileextension, starttime, endtime)
diagspace = comm.DiagSpace(common.spatialgrid, x_fourier=common.x_local, y_fourier=common.y_local,
                           z_fourier=False, xrange=xrange, yrange=yrange, zrange=zrange, xavg=args.xavg,
                           yavg=args.yavg, zavg=args.zavg)
outputslice = sli.MomFieldSlice(common, args.var, args.specname, diagspace)

outputslice.generate_timeseries()
outputslice.write_to_numpy("slice_{}_{}".format(args.specname, fileextension))
outputslice.write_to_ascii("slice_{}_{}.dat".format(args.specname, fileextension))



""" plot_srcmom.py: Plotting for the binary files of the source moment diagnostic"""


import matplotlib.pyplot as plt
from .baseplot import Plotting
from ..data.srcmom_data import SrcmomSeries
from ..utils.averages import mytrapz


class PlotSrcmomAverage(Plotting):
    """Plot time averages of the source profiles

    :param srcmomseries: List of SrcmomData objects we want to plot
    """

    def __init__(self, srcmomseries):
        super().__init__()

        # Lists for the plotted structures
        self.leglist = []
        self.figlist = None
        self.axlist = None
        # Data structures used in the different plots
        self.src = srcmomseries
        for src in self.src:
            src.generate_timeseries()

    def createfigs(self):
        numfig = len(self.src[0].srcmoms)
        self.figlist = [plt.figure() for _ in range(numfig)]
        self.axlist = [fig.add_subplot(111) for fig in self.figlist]
        for srcav in self.src:
            self.plot_srcmoment(self.axlist, srcav)
        for ax in self.axlist:
            ax.legend()

    def plot_srcmoment(self, axlist, srcav):
        for isrc, srcdat in enumerate(srcav.srcmoms):
            axlist[isrc].plot(srcav.cm.spatialgrid.x_a, mytrapz(srcdat, srcav.timearray),
                              label=srcav.momtype + " " + srcav.srctype)
            axlist[isrc].set_xlabel(r'$x/a$')
            axlist[isrc].set_title(srcav.cm.specnames[isrc])


class PlotSrcmomxt(Plotting):
    """ Plot x-t colormaps of the source moments"""

    def __init__(self, srcmomaverage):
        super().__init__()



""" Module with the container objects for simulation data"""
# -*- coding: utf-8 -*-

import numpy as np
from .par_io import Parameters
from .geom import Geometry
from ...data.nrgdata import NrgFile
from ...data.data import Data
from ...data.profiledata import ProfileData

class Run:
    """ Class containing basic information on a simulation data set"""

    def __init__(self, in_folder, fileextension):
        self.in_folder = in_folder
        self.fileextension = fileextension
        # parameters is read always from the ascii file
        params = Parameters()
        print('fileextension')
        params.read_pars(self.in_folder + '/parameters{}'.format(fileextension))

        self.pars = params.pardict
        self.pnt = params.asnamedtuple()
        # Determine if we are in one of the global versions
        self.x_local = self.pars.get('x_local', True)  # By default, x_local = T is not written
        self.y_local = self.pars.get('y_local', True)  # By default, y_local = T is not written
        self.is3d = not self.x_local and not self.y_local
        self.x_global = not self.x_local and self.y_local
        self.flux_tube = self.x_local and self.y_local
        self.nonlinear = self.pars.get('nonlinear')
        self.electromagnetic = (self.pars.get('beta', False) != 0)
        self.bpar = self.pars.get("bpar", False)

        # which output we are handling
        self.is_h5 = self.pars.get('write_h5', False)
        self.is_adios = self.pars.get('write_hac', False)

        # Here we patch the extension adding the right suffix
        if self.is_h5:
            self.fileextension = fileextension + ".h5"
        elif self.is_adios:
            self.fileextension = fileextension + ".bp"

        # each run contians its own geometry info, for e.g. change of resolution
        self.geometry = Geometry(self)

        # each run contains also species. For changes or just for completeness
        self.specnames = params.specnames

        # spatial grids
        self.spatialgrid = SpatialGrid(self.pnt, self.x_local, self.y_local)

        self.vspgrid = VspGrid(self.pnt)

        if not self.flux_tube:
            self.profilesdata = ProfileData(self.pnt, self.is_h5,
                                            self.specnames, self.fileextension,
                                            self.in_folder)


class Simulation:
    """ The simulations class embeds the Run class, meaning a simulation is a collection
        of runs. Aiming at something where a single run can be handled by itself
        for cases to be used not in the GUI """

    def __init__(self, in_folder=None, out_folder=None, extensions=None):
        self.in_folder = in_folder
        self.extensions = extensions
        self.out_folder = out_folder
        self.starttime = -1
        self.endtime = -2
        self.stepping = 1
        self.max_steps = 'all'
        self.data = Data()
        self.specnames = None
        self.run = []
        self.nrgdata = []

    def prepare(self):
        """ Sort extensions in time ascending

         We dont really need it, but looks better on the GUI
        """

        first_time = []

        if self.run:
            self.run.clear()
        if self.nrgdata:
            self.nrgdata.clear()

        for i_ext, ext in enumerate(self.extensions):
            # there is a lot to cleanup
            # run contains each of the restart we want to process
            act_run = Run(self.in_folder, ext)
            self.extensions[i_ext] = act_run.fileextension
            self.run.append(act_run)

        # calling Set fills out the nrg info only
        for irun, ext in enumerate(self.extensions):
            # list all nrg
            nrg = NrgFile(self.in_folder+ '/nrg{}'.format(self.extensions[irun]), self.run[irun])
            self.nrgdata.append(nrg)
            nrg.generate_timeseries()
            first_time.append(nrg.timearray[0])

        self.run = [x for _, x in sorted(
                zip(sorted(range(len(first_time)), key=lambda k: first_time[k]), self.run))]
        self.extensions = [x for _, x in sorted(
                zip(sorted(range(len(first_time)), key=lambda k: first_time[k]), self.extensions))]
        self.nrgdata = [x for _, x in sorted(
                zip(sorted(range(len(first_time)), key=lambda k: first_time[k]), self.nrgdata))]

        self.data.load_in(self.run[0], self.extensions)
        self.specnames = self.run[0].specnames


class SpatialGrid:
    """ The numerical grid based on the parameters of a GENE run

    :param pnt: NamedTuple of the run parameters
    :param x_local: Is the run local in radial direction?
    :param y_local: Is the run local in "binormal" direction?
    """

    def __init__(self, pnt, x_local, y_local):
        try:
            kx_center = pnt.kx_center
        except AttributeError:
            kx_center = 0

        self._calc_kxgrid(x_local, y_local, pnt.lx, pnt.nx0, kx_center)
        self._calc_xgrid(x_local, pnt.lx, pnt.nx0, pnt.rhostar, pnt.x0)
        self._calc_kygrid(x_local, y_local, pnt)
        self._calc_ygrid(x_local, y_local, pnt)
        self.z = np.array([-np.pi + 2*np.pi*k/pnt.nz0 for k in range(pnt.nz0)], dtype=np.float64)
        self.nx0 = pnt.nx0
        if not y_local:
            self.ny0 = pnt.ny0
        self.nz0 = pnt.nz0


    def _calc_xgrid(self, x_local, lx, nx0, rhostar, x0):
        if x_local:
            xgrid = np.fft.fftfreq(nx0, d=1.0/lx)
            self.x = np.fft.fftshift(xgrid)
        else:
            self.x = np.array([-lx/2. + lx/(nx0 - 1)*i for i in range(0, nx0)])
            self.x_a = self.x*rhostar + x0

    def _calc_kxgrid(self, x_local, y_local, lx, nx0, kx_center=0):

        self.kxmin = 2*np.pi/lx

        if x_local:
            if y_local:
                kxgridp = [self.kxmin*i for i in range(int((nx0 + 1)/2) + (1 - nx0%2))]
                kxgridn = [self.kxmin*i for i in range(int(-(nx0 - 1)/2), 0)]
                # positive kx modes
                self.kx_pos = np.array(kxgridp) + kx_center
                self.kx_fftorder = np.array(kxgridp + kxgridn, dtype=np.float64) + kx_center
                # ordered kx array
                self.kx = np.array(kxgridn + kxgridp, dtype=np.float64) + kx_center
            else:
                self.kx = np.array([self.kxmin*i for i in range(nx0)])
        else:
            self.kx_fftorder = 2*np.pi*np.fft.fftfreq(nx0, d=lx/nx0)
            self.kx = np.fft.ifftshift(self.kx_fftorder)
            self.kx_pos = self.kx_fftorder[:int(nx0/2 + 1)]

    def _calc_ygrid(self, x_local, y_local, pnt):
        ly = 2*np.pi/pnt.kymin
        if y_local:
            self.y = np.fft.fftshift(np.fft.fftfreq(2*pnt.nky0, d=1/ly))
        else:
            if x_local:
                self.y = np.array([-ly/2. + ly/pnt.nky0*i for i in range(0, pnt.nky0)])
            else:
                self.y = np.array([ly/pnt.ny0*i for i in range(0, pnt.ny0)])

    def _calc_kygrid(self, x_local, y_local, pnt):
        if y_local:
            self.ky = np.array([pnt.kymin*(j + pnt.ky0_ind) for j in range(pnt.nky0)],
                               dtype=np.float64)
        else:
            ly = 2*np.pi/pnt.kymin
            if x_local:
                self.ky = 2*np.pi*np.fft.ifftshift(np.fft.fftfreq(pnt.nky0, d=ly/pnt.nky0))
            else:
                self.ky = np.fft.fftfreq(pnt.ny0)*pnt.ny0*pnt.kymin

    def kx_grid_slice(self, kxslice):
        """Get a full kx grid in FFT order from a slice for positive kx"""
        sliced_kxpos = self.kx_pos[kxslice]
        if self.nx0 % 2 == 1:
            sliced_kxneg = -sliced_kxpos[-1::-1]
        else:
            if kxslice.stop is None:  # Nyquist mode must be ignored
                sliced_kxneg = -sliced_kxpos[-2::-1]
            else:
                sliced_kxneg = -sliced_kxpos[-1::-1]
        if kxslice.start is None:
            sliced_kxneg = sliced_kxneg[:-1]
        sliced_kx = np.concatenate((sliced_kxpos, sliced_kxneg))
        return sliced_kx


class VspGrid:
    def __init__(self, pnt):
        self._calc_vgrid(pnt.lv, pnt.nv0)
        if pnt.mu_grid_type == 'gau_lag':
            self._calc_mugrid_gaulag(pnt.lw, pnt.nw0)
        else:
            raise NotImplementedError("This mu grid is not implemented")

    def _calc_vgrid(self, lv, nv):
        dv = 2.0*lv/(nv-1.0)
        self.vp = np.zeros(nv)
        for iv in range(0, nv):
            self.vp[iv] = -lv + iv*dv

    def _calc_mugrid_gaulag(self, lw, nw):
        self.mu, dmu = np.polynomial.laguerre.laggauss(nw)
        dmu = dmu * np.exp(self.mu)
        fac = lw / np.sum(dmu)
        self.mu = self.mu*fac



from .fieldlib import FieldFile
from .momlib import MomFile
from .nrgdata import NrgFile
from .profile_data import ProfileFile
from .srcmom_data import SrcmomFile


class RunDataFiles(object):
    """ Collection of the various output file objects of a GENE run

    The purpose of this class is to avoid rereading large binary files several times
    """

    def __init__(self, common):
        self.cm = common
        self.filedict = {"field": None, "nrg": None}
        for spec in common.specnames:
            self.filedict.update({"mom_{}".format(spec): None, "profile_{}".format(spec): None,
                                  "srcmom_{}".format(spec): None})

    def get_fileobject(self, name):
        if not self.filedict.get(name):
            if name == "field":
                self.filedict["field"] = FieldFile(
                    "field{}".format(self.cm.fileextension), self.cm)
            elif name == "nrg":
                self.filedict["nrg"] = NrgFile("nrg{}".format(self.cm.fileextension), self.cm)
                self.filedict["nrg"].generate_timeseries()
            elif name.startswith("mom_"):
                self.filedict[name] = MomFile("{}{}".format(name, self.cm.fileextension),self.cm)
            elif name.startswith("profile_"):
                self.filedict[name] = ProfileFile("{}{}".format(name, self.cm.fileextension),self.cm)
            elif name.startswith("srcmom_"):
                self.filedict[name] = SrcmomFile("{}{}".format(name, self.cm.fileextension),self.cm)

            else:
                raise RuntimeError("Invalid file object identifier")
        return self.filedict[name]



import json, bson
import numpy as np
from pyrokinetics import Pyro,template_dir
from pyrokinetics.databases.imas import pyro_to_imas_mapping
import idspy_toolkit as idspy
from idspy_dictionaries import ids_gyrokinetics_local as gkids
from pathlib import Path
import os

# 
def convert_to_json(obj,separate_real_imag = False):
    """
    This function to recursively goes through GyrokineticsLocal class, and writes to json compatible dictionary

    Parameters
    ----------
    obj : GyrokineticsLocal
        GyrokineticsLocal object with data loaded
    separate_real_imag : bool
        If true, move real and imaginary parts to separate dictionary keys. 
        If False, encode complex np.array

    Returns
    -------
    json compatible dictionary.

    """
    
    if '__dataclass_fields__' in dir(obj):
        tmpdict = {}
        for item in obj.__dataclass_fields__.keys():
            
            value =  eval(f"obj.{item}")
            if separate_real_imag and isinstance(value, np.ndarray) and 'complex' in str(value.dtype).lower():
                #print(value)
                tmpdict[item + "_real"] = convert_to_json(value.real)
                tmpdict[item + "_imaginary"] = convert_to_json(value.imag)
            else:
                tmpdict[item] = convert_to_json(value)
        return tmpdict
    elif isinstance(obj, np.ndarray):
        if 'complex' in str(obj.dtype).lower():
                return dict(
                    __ndarray_tolist_real__=obj.real.tolist(),
                    __ndarray_tolist_imag__=obj.imag.tolist(),
                    dtype=str(obj.dtype),
                    shape=obj.shape,
                )
        else:
            return obj.tolist()
    elif isinstance(obj, dict):
        return {key: convert_to_json(value) for key, value in obj.items()}
    elif isinstance(obj, (list, tuple)):
        return [convert_to_json(item) for item in obj]
    else:
        return obj

def update_key_values(data, mod_key, new_value):
    '''
    Module to iteratively scan for entries in a dictionary or sub dictionaries or sublists
    with a specific key and repalce it with a given value
    '''
    
    if isinstance(data, dict):
        for key, value in data.items():
            if key == mod_key:
                data[key] = new_value
            else:
                update_key_values(value, mod_key, new_value)
    elif isinstance(data, list):
        for item in data:
            update_key_values(item, mod_key, new_value)


def prune_imas_gk_dict(gk_dict, pyro, linear):
    ''' Remove 4d files for linear and non-linear runs  
    Also set value of max_repr_length using input values 
    '''

    if linear: # If linear, drop entries in ['linear']['wavevector'][0]['eigenmode'][i]['fields'][key] 
        if gk_dict['non_linear']['fields_4d'] is not None:   
            assert gk_dict['non_linear']['fields_4d']['phi_potential_perturbed_norm']==[], "phi_potential_perturbed_norm field in non_linear, fields_4d is not empty"

        keys_list = ['phi_potential_perturbed_norm','a_field_parallel_perturbed_norm','b_field_parallel_perturbed_norm']
        if gk_dict['linear']['wavevector'] !=[]: 
            for i in range(len(gk_dict['linear']['wavevector'][0]['eigenmode'])): ## For each particle species, delete fields
                if gk_dict['linear']['wavevector'][0]['eigenmode'][0]['fields']: 
                    for key in keys_list:
                        gk_dict['linear']['wavevector'][0]['eigenmode'][i]['fields'][key]=None

    else: # If non-linear, drop  ['non_linear']['fields_4d]
        assert (gk_dict['linear']['wavevector']==[]),"wavevector field in linear is not empty"

        gk_dict['non_linear']['fields_4d']=None

    ## Setup max_repr_length
    if pyro._gk_code=='GENE':
        prec = pyro.gk_input.data["info"]["PRECISION"] 
        prec_dict = {'SINGLE':32, 'DOUBLE':64}
        max_repr_length = prec_dict[prec]

        update_key_values(gk_dict, 'max_repr_length', max_repr_length)
    
    return gk_dict

def create_gk_dict_with_pyro(fname,gkcode):
    '''
    Create gyrokinetics dictionary to be upload to database
    '''

    assert gkcode in ['GENE','CGYRO','TGLF','GS2','GX'], "invalid gkcode type %s"%(gkcode)
    
    try: 
        pyro = Pyro(gk_file=fname, gk_code=gkcode)
        linear = not pyro.numerics.nonlinear

        if gkcode=='TGLF':   
            quasi_linear = pyro.numerics.nonlinear
            linear = True
        else:      
            quasi_linear = False 

        if linear: 
            pyro.load_gk_output(load_fields=True)
        else: # Loading fields for non-linear runs can take too long, so do not read them 
            pyro.load_gk_output(load_fields=False)

        gkdict = gkids.GyrokineticsLocal()
        idspy.fill_default_values_ids(gkdict)
        gkdict = pyro_to_imas_mapping(
                pyro,
                comment=f"Computing IMAS for %s"%(gkcode),
                ids=gkdict
            )
        
        json_data = convert_to_json(gkdict)

        json_data = prune_imas_gk_dict(json_data, pyro, linear)

        ## Confirm IMAS size is less than 2MB
        bson_data = bson.BSON.encode(json_data)
        imas_size = len(bson_data)/1e6 # IMAS size in Megabytes
        if imas_size  > 2.0 : 
            print("Size of IMAS dict: ",imas_size,"MB")
            print("IMAS size is larger than 2MB. Need to check IMAS content")
            raise SystemError
    
    except Exception as e: 
        print(e)
        raise SystemError
    
    return json_data,quasi_linear



# -*- coding: utf-8 -*-

import tkinter as tk
from tkinter import END
from tkinter import Toplevel, BOTTOM, BOTH, TOP

import h5py
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
import numpy as np

import ..utils.aux_func as aux_func
from .baseplot import Plotting
from .diagnostic import Diagnostic


class DiagChivsGradientgene3d(Diagnostic):
    def __init__(self, avail_vars=None, specnames=None):
        super().__init__()
        self.name = 'Chi vs. Gradient'
        self.tabs = ['GENE3D']
        self.help_txt = """Plots chi vs. gradient
                        \nx ind : comma separated radial interval in terms of minor radious (e.g. 
                        0.4,0.6)
                        \ndefault all range
                        \nSave h5 : save hdf5 file (def. True)"""

        self.avail_vars = avail_vars
        self.specnames = specnames

        self.opts = {"spec": {'tag': "Spec", 'values': self.list_specnames()},
                     "xind": {'tag': "x ind", 'values': None},
                     "save_h5": {"tag": "Save h5", "values": [True, False]}}

        self.set_defaults()

        self.options_gui = Diagnostic.OptionsGUI()

    def set_options(self, run_data, species):

        # radial position, unset or -1 means all
        xind = self.opts['xind']['value']
        if not xind or xind == -1:
            self.xind = 0, 100  # this will choose the minimum and maximum values
        else:
            self.xind = xind.split(',')
        self.xind_min, self.xind_max = aux_func.find_nearest_points(run_data.spatialgrid.x_a,
                                                                    float(self.xind[0]),
                                                                    float(self.xind[1]))

        class ChiStep:
            def __init__(self, specnames):
                for spec in specnames:
                    setattr(self, spec, self.__Chiprofiles())

            class __Chiprofiles:
                def __init__(self):
                    self.omt = []
                    self.omn = []
                    self.chi_es = []

        self.run_data = run_data

        self.geom = run_data.geometry

        self.chi_step = ChiStep(self.specnames)

        self.get_needed_vars(['mom'])

        self.get_spec_from_opts()

        return self.needed_vars

    def execute(self, data, step):

        for ispec, spec in enumerate(self.specnames):

            chi_step = getattr(self.chi_step, spec)

            n = np.average(getattr(getattr(data, 'mom_' + spec), "n")(step.time, step.mom),
                           weights=self.geom.jacobian.T, axis=(1, 2))
            T_par = np.average(getattr(getattr(data, 'mom_' + spec), "T_par")(step.time, step.mom),
                               weights=self.geom.jacobian.T, axis=(1, 2))
            T_per = np.average(getattr(getattr(data, 'mom_' + spec), "T_per")(step.time, step.mom),
                               weights=self.geom.jacobian.T, axis=(1, 2))
            Q_es = np.average(getattr(getattr(data, 'mom_' + spec), "Q_es")(step.time, step.mom),
                              weights=self.geom.jacobian.T, axis=(1, 2))
            T_total = 1/3*T_par + 2/3*T_per

            T0 = self.run_data.profilesdata.T0s[:, ispec]/self.run_data.pars["temp" + spec]/ \
                 self.run_data.pars["Tref"]
            N0 = self.run_data.profilesdata.n0s[:, ispec]/self.run_data.pars["dens" + spec]/ \
                 self.run_data.pars["nref"]

            Temp = T0 + T_total*self.run_data.pnt.rhostar*self.run_data.pnt.minor_r
            Dens = N0 + n*self.run_data.pnt.rhostar*self.run_data.pnt.minor_r

            omt = -np.gradient(np.log(Temp),
                               self.run_data.spatialgrid.x_a)/self.run_data.pnt.minor_r
            omn = -np.gradient(np.log(Dens),
                               self.run_data.spatialgrid.x_a)/self.run_data.pnt.minor_r

            chi_step.chi_es.append(np.mean(Q_es[self.xind_min:self.xind_max + 1]/(
                        Dens[self.xind_min:self.xind_max + 1]*Temp[self.xind_min:self.xind_max + 1]
                        *omt[self.xind_min:self.xind_max + 1])))
            chi_step.omt.append(np.mean(omt[self.xind_min:self.xind_max + 1]))
            chi_step.omn.append(np.mean(omn[self.xind_min:self.xind_max + 1]))

    def plot(self, time_requested, output=None, out_folder=None):

        if output:
            output.info_txt.insert(END, "Chi vs. gradient:\n")

        self.plotbase = Plotting()

        x_lbl = r'$t\; [c_{ref}/L_{ref}]$'
        text_x = ', Avg. over x = [' + '{:.1f}'.format(
                self.run_data.spatialgrid.x_a[self.xind_min]) + ',' + '{:.1f}'.format(
                self.run_data.spatialgrid.x_a[self.xind_max - 1]) + ']  '

        # time averaged values
        for spec in self.specnames:

            spec_chi = getattr(self.chi_step, spec)

            fig = plt.figure(figsize=(15, 10), dpi=100)
            fig.suptitle(r"${}    $".format(spec) + text_x)

            ax_1 = fig.add_subplot(2, 2, 1)
            ax_2 = fig.add_subplot(2, 2, 2)
            ax_3 = fig.add_subplot(2, 2, 3)
            ax_4 = fig.add_subplot(2, 2, 4)

            chi_es = getattr(spec_chi, "chi_es")
            omt = getattr(spec_chi, "omt")
            omn = getattr(spec_chi, "omn")

            ax_1.plot(time_requested, omt)
            ax_1.set_xlabel(x_lbl)
            ax_1.set_ylabel(r"$a/L_{T}$")

            ax_2.plot(time_requested, omn)
            ax_2.set_xlabel(x_lbl)
            ax_2.set_ylabel(r"$a/L_{n}$")

            ax_3.plot(time_requested, chi_es)
            ax_3.set_xlabel(x_lbl)
            ax_3.set_ylabel(r"$<\chi_{es}>_{V}$")

            ax_4.plot(omt, chi_es)
            ax_4.set_xlabel(r'$a/L_{T}$')
            ax_4.set_ylabel(r"$<\chi_{es}>_{V}$")

            # this was an idea, but not sure if it is worh it
            window = Toplevel()
            tk.Tk.wm_title(window, "Chi vs. Gradient")
            ShowThis(window, fig)

        if self.opts['save_h5']['value']:
            file = h5py.File(out_folder/'chi_gradient_vs_time.h5', 'w')
            file["/time"] = time_requested
            file["/omt"] = omt
            file["/omn"] = omn
            file["/chi_es"] = chi_es
            file.close()


class ShowThis:
    def __init__(self, window, f):
        canvas = FigureCanvasTkAgg(f, window)
        canvas.draw()
        canvas.get_tk_widget().pack(side=BOTTOM, fill=BOTH, expand=True)

        try:
            from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk
            toolbar = NavigationToolbar2Tk(canvas, window)
            toolbar.update()
            canvas._tkcanvas.pack(side=TOP, fill=BOTH, expand=True)
        except:  # TODO Really swallow all exceptions without saying something?
            pass
